#!/usr/bin/env python
# ===================================================================
# ScriptName:   tgt
# Purpose:      Class to hold and process target areas
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/07
# Revision:     v1.2
# Last Change:  2024/09/02: fixed area name not considered when preparing targets
#               2024/08/19: added virtual map creation, nav preparation, target export
#               2024/08/16: added PACEArea child of TargetArea
#               2024/08/12: implemented local optimization
#               2024/08/09: added method to change target area for point
#               2024/08/07: separated from old SPACEtomo_tgt.py
# ===================================================================

import json
import time
from pathlib import Path
import numpy as np
from scipy.optimize import minimize
from scipy.cluster.vq import kmeans2

from SPACEtomo.modules.ext import calcScore_cluster, calcScore_point
from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log


class Targets:
    """Class to hold all target areas of lamella with functionality to redistribute points among target areas and facilitate access on the points level."""

    def __init__(self, map_dir, map_name, tgt_params, map_dims, map_pix_size=None) -> None:
        self.map_dir = Path(map_dir)
        self.map_name = map_name            # needed for file names
        self.map_dims = map_dims            # needed to check bounds
        self.map_pix_size = map_pix_size    # needed for meta_data
        self.tgt_params = tgt_params        # needed for rec_dims and target score calculation

        # Get camera dims
        self.rec_dims = np.array(tgt_params.weight.shape)

        self.areas = []
        self.settings = None

        # Set class attributes for areas
        TargetArea.tgt_params = self.tgt_params
        TargetArea.rec_dims = self.rec_dims



    def addTarget(self, coords, new_area=False):
        """Checks conditions and adds new target to closest target area."""

        # Check if coords are out of bounds
        if not self.checkBounds(coords):
            log("WARNING: Can't add target. Too close to the edge!")
            return False
        
        # Create target areas to allow manual picking without auto run
        if not self.areas:
            self.areas.append(TargetArea())

        # Check if coords are too close to existing point
        closest_point_id, in_range = self.getClosestPoint(coords, np.min(self.rec_dims))
        if in_range:
            log("WARNING: Target is too close to an existing target! It will not be added.")
            return False
        
        if new_area:
            # Create new area only if last area has points
            if self.areas[-1]:
                self.areas.append(TargetArea())
            closest_area = -1

        else:
            # Figure out which target area tracking targets is closest
            closest_area = self.getClosestArea(coords)

        # Add point
        self.areas[closest_area].addPoint(coords)
        log("NOTE: Added new target!")
        return True

    def addGeoPoint(self, coords):
        """Similar to addTargets for geo points with different checks."""

        # Check if coords are out of bounds
        if not self.checkBounds(coords):
            log("WARNING: Can't add point. Too close to the edge!")
            return False
        
        # No geo points without any targets selected
        if not self.areas:
            log("WARNING: Can't add geo point without any selected targets!")
            return False

        # Check if coords are too close to existing point
        closest_point_id, in_range = self.getClosestPoint(coords, np.min(self.rec_dims))
        if in_range:
            log("WARNING: Point is too close to an existing target! It will not be added.")
            return False
        
        # Add geo point to all target areas
        for area in self.areas:
            area.addPoint(coords, geo=True)
        log("NOTE: Added geo point!")
        return True
    
    def movePointToArea(self, id, new_area_id):
        """Moves point from old target area to new target area."""

        old_area_id = id[0]
        old_point_id = id[1]
    
        # Get point data
        coords = self.areas[old_area_id].points[old_point_id]
        score = self.areas[old_area_id].scores[old_point_id]

        # Add to new area
        self.areas[new_area_id].addPoint(coords, score=score)

        # Remove from old area (and area if no points remaining)
        self.areas[old_area_id].removePoint(old_point_id)
        if len(self.areas[old_area_id]) == 0:
            self.areas.pop(old_area_id)
    
    def getAllPoints(self, include_geo=False):
        """Creates list of all points and list of original ids."""
        
        points = np.empty([0, 2])
        ids = []
        for a, area in enumerate(self.areas):
            points = np.vstack([points, area.points])
            ids += [[a, p] for p in range(len(area.points))]        # keep track of area and point ids for each point
        return points, ids

    def getClosestPoint(self, coords, threshold):
        """Finds closest point and checks if distance within threshold."""

        points, point_ids = self.getAllPoints()
        # Check for point within range
        if len(points) > 0:
            closest_id = np.argmin(np.linalg.norm(points - coords, axis=1))
            return point_ids[closest_id], np.linalg.norm(points[closest_id] - coords) < threshold
        else:
            return None, False

    def getClosestArea(self, coords):
        """Finds area with closest tracking point."""

        if len(self.areas) > 1:
            # Make list of center points and exclude empty areas
            track_points = [target_area.center for target_area in self.areas if target_area.center is not None]
            # Get ID of closest center point
            closest_area = np.argmin(np.linalg.norm(track_points - coords, axis=1))
        else:
            closest_area = 0
        return closest_area
    
    def checkBounds(self, coords):
        """Checks if coords are out of bounds of map."""

        if not self.rec_dims[0] / 2 <= coords[0] < self.map_dims[0] - self.rec_dims[0] / 2 or not self.rec_dims[1] / 2 <= coords[1] < self.map_dims[1] - self.rec_dims[1] / 2:
            return False
        return True

    def resetGeo(self):
        for area in self.areas:
            area.geo_points = np.empty([0, 2])

    def mergeAreas(self):
        """Merge all target areas."""

        # Check if more than area exist
        if len(self.areas) > 1:

            # Create new area
            merged_area = TargetArea()

            # Extract points from all previous areas
            for area in self.areas:
                for point, score in zip(area.points, area.scores):
                    merged_area.addPoint(point, score=score)

                for point in area.geo_points:
                    # Check if geo point was already added to merged_area
                    if np.any([np.array_equal(point, geo_point) for geo_point in merged_area.geo_points]):
                        continue
                    merged_area.addPoint(point, geo=True)

            # Keep only merged area
            self.areas = [merged_area]

    def splitArea(self, area_num=2):
        """Use kmeans clustering to split target area."""

        # Merge previous areas
        self.mergeAreas()

        # Cluster all points
        centroids, labels = kmeans2(self.areas[0].points, area_num, minit="++")
        counts = np.bincount(labels)
        log(f"NOTE: Split targets into {len(counts)} areas with {', '.join(counts.astype(str))} targets, respectively.")

        # Create new areas and add points according to their label
        new_areas = []
        for point_id, area_id in enumerate(labels):
            while area_id >= len(new_areas):
                new_areas.append(TargetArea())
            new_areas[area_id].addPoint(self.areas[0].points[point_id], score=self.areas[0].scores[point_id])

        # Add geo point to all target areas
        for geo_point in self.areas[0].geo_points:
            for area in new_areas:
                area.addPoint(geo_point, geo=True)

        # Overwrite target areas
        self.areas = new_areas

    def splitAreaManual(self):
        """Split targets by distance to existing tracking points."""

        remove_list = []
        for area_id, area in enumerate(self.areas):
            for point_id, (point, score) in enumerate(zip(area.points, area.scores)):
                # Skip tracking target
                if point_id == 0: continue

                # Find closest tracking point
                closest_area_id = self.getClosestArea(point)

                # If closest tracking point is not current area
                if closest_area_id != area_id:
                    # Move point from current to closest area
                    self.areas[closest_area_id].addPoint(point, score=score)
                    remove_list.append((area_id, point_id))

        # Remove points that were moved (starting from highest ids)
        for area_id, point_id in sorted(remove_list, key=lambda x: x[1], reverse=True):
            self.areas[area_id].removePoint(point_id)

    def loadAreas(self):
        """Loads json data for all point files."""

        # Find all point files
        point_files = sorted(self.map_dir.glob(self.map_name + "_points*.json"))

        if len(point_files) > 0:
            # Instantiate target area for each file
            self.areas = []
            for file in point_files:
                area = TargetArea(file=file)
                # Only load non-empty point files
                if len(area) > 0:
                    self.areas.append(area)
            log("NOTE: Loaded target coordinates from file.")
        else:
            log("NOTE: No target coordinates found.")
            self.areas = []

        # Check if settings were loaded from file
        if TargetArea.settings:
            self.settings = TargetArea.settings

    def exportTargets(self, settings=None):
        """Exports all target areas to json files."""

        # Check for existing point files and delete them
        point_files = sorted(self.map_dir.glob(self.map_name + "_points*.json"))
        if len(point_files) > 0:
            log("NOTE: Previous targets were deleted.")
            for file in point_files:
                # Delete file on path
                file.unlink()

        meta_data = {"pix_size": self.map_pix_size, "img_size": self.map_dims, "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}

        if self.areas:
            for a, area in enumerate(self.areas):
                if not area: continue   # Don't save empty areas
                file = self.map_dir / (self.map_name + f"_points{a}.json")
                area.exportToJson(file, settings, meta_data)
        else:
            # Write empty points file to ensure empty targets file is written and map is considered processed
            with open(self.map_dir / (self.map_name + "_points.json"), "w") as f:
                json.dump({"points": [], "scores": [], "geo_points": [], "meta_data": meta_data}, f)    

        # Update settings
        if settings:
            self.settings = settings

    def __bool__(self):
        """Return False if no points exist."""

        return bool(self.getAllPoints()[1])     # Have to check list of ids since points is a numpy array
    
    def __len__(self):
        return len(self.getAllPoints()[0])
    
    def __str__(self):
        text = "Target areas:\n"
        for a, area in enumerate(self.areas):
            text += f"Area {a}:\t{len(area)} points\n"
        return text

class TargetArea:
    """Class to hold all data for one target area."""

    # Class attributes
    tgt_params = None
    rec_dims = None

    settings = None

    def __init__(self, file=None) -> None:
        self.points = np.empty([0, 2])
        self.scores = np.empty(0)
        self.geo_points = np.empty([0, 2])

        # Dict to hold class dependent scores
        self.class_scores = {}

        if file:
            self.loadFromJson(file)

        if len(self.points) > 0:
            self.center = self.points[0]
        else:
            self.center = None

    def loadFromJson(self, file):
        """Reads targets from json file."""

        # Load json data
        with open(file, "r") as f:
            area = json.load(f, object_hook=utils.revertArray)

        if np.any(area["points"]):
            self.points = area["points"]
        if np.any(area["scores"]):
            self.scores = area["scores"]
        if np.any(area["geo_points"]):
            self.geo_points = area["geo_points"]
        if "settings" in area.keys():
            TargetArea.settings = area["settings"]

    def exportToJson(self, file, settings=None, meta_data=None):
        """Writes targets to json file."""

        # Convert to dictionary
        output_dict = {"points": self.points, "scores": self.scores, "geo_points": self.geo_points}
        if settings:
            output_dict["settings"] = settings
            # Update class attribute
            TargetArea.settings = settings
        if meta_data:
            output_dict["meta_data"] = meta_data

        # Write to file
        with open(file, "w") as f:
            json.dump(output_dict, f, indent=4, default=utils.convertArray)

    def addPoint(self, coords, geo=False, score=100):
        """Adds new point at coords."""

        if not geo:
            self.points = np.vstack([self.points, coords])
            self.scores = np.append(self.scores, [score])
            if self.center is None:
                self.center = coords
        else:
            self.geo_points = np.vstack([self.geo_points, coords])

    def removePoint(self, id):
        """Removes point by id."""

        self.points = np.delete(self.points, id, axis=0)
        self.scores = np.delete(self.scores, id, axis=0)

        # Update center
        self.center = self.points[0]

    def makeTrack(self, id):
        self.points[0: id + 1] = np.roll(self.points[0: id + 1], shift=1, axis=0)

        # Update center
        self.center = self.points[0]

    def getPointInfo(self, id):
        """Gets information about point by id."""

        dist = np.linalg.norm(self.points[id] - self.center)
        return self.scores[id], dist
    
    def getClassScores(self, cat, mask):
        """Calculates scores for single classes."""

        self.class_scores[cat] = np.empty(0)
        for point in self.points:
            score = calcScore_point(point, mask, np.zeros(mask.shape), 0, self.tgt_params.weight, self.tgt_params.edge_weights)
            self.class_scores[cat] = np.append(self.class_scores[cat], [score])
    
    def optimizeLocally(self, point_id, mask, penalty_mask=None):
        """Use mask to optimize target position locally."""

        # Check if class attributes were set
        if not self.tgt_params or self.rec_dims is None:
            log("ERROR: Target params have not been set for target area!")
            return
        
        # Check if penalty mask was given
        if not penalty_mask:
            penalty_mask = np.zeros(mask.shape)
        
        # Get initial coords and score
        point = self.points[point_id, :]
        score = calcScore_point(point, mask, penalty_mask, self.tgt_params.penalty, self.tgt_params.weight, self.tgt_params.edge_weights)
        log(f"Initial score: {round(score, 3)}")

        # Call optimization
        start_offset = np.zeros(2)
        offset_raw = minimize(calcScore_cluster, start_offset, args=(point[np.newaxis, :], mask, penalty_mask, self.tgt_params.penalty, self.tgt_params.weight, self.tgt_params.edge_weights), method="nelder-mead", bounds=((-0.1, 0.1), (-0.1, 0.1)))
        offset_pixel = offset_raw.x * max(self.rec_dims) * 10
        point += offset_pixel

        # Calculate new score
        log(f"Target was moved by {round(np.linalg.norm(offset_pixel))} pixels.")
        score = calcScore_point(point, mask, penalty_mask, self.tgt_params.penalty, self.tgt_params.weight, self.tgt_params.edge_weights)
        log(f"New score: {round(score, 3)}")

        # Update point
        self.points[point_id] = point
        self.scores[point_id] = score
    
    def __bool__(self):
        return bool(len(self.points))

    def __len__(self):
        return len(self.points)


class PACEArea(TargetArea):
    """Target area with additional functions to create PACEtomo tgts file."""

    model = None
    imaging_params = None

    def __init__(self, file, model, imaging_params, buffer) -> None:
        super().__init__(file)

        PACEArea.model = model
        PACEArea.imaging_params = imaging_params

        self.pix_size = self.model.pix_size
        self.map_buffer = buffer

        self.stageCoords()
        self.specimenCoords()

    def getClassScores(self, cat, mask):
        raise NotImplementedError("This method is not implemented for PACEArea!")
    
    def optimizeLocally(self, point_id, mask, penalty_mask=None):
        raise NotImplementedError("This method is not implemented for PACEArea!")

    def stageCoords(self):
        """Makes list of points in stage coords."""

        self.scaleCoordsBuffer()
        self.points_stage = np.array([self.map_buffer.px2stage(np.flip(point)) for point in self.points])
        if self.center is not None:
            self.center_stage = self.points_stage[0]
        else:
            self.center_stage = None
        self.geo_points_stage = np.array([self.map_buffer.px2stage(np.flip(point)) for point in self.geo_points])

    def specimenCoords(self):
        """Makes list of points in relative specimen coords."""

        if self.center_stage is not None:
            self.points_ss = np.array([self.imaging_params.s2ss_matrix @ (point - self.center_stage) for point in self.points_stage])
            self.geo_points_ss = np.array([self.imaging_params.s2ss_matrix @ (point - self.center_stage) for point in self.geo_points_stage])
        else:
            log("WARNING: Can't convert to relative specimen coordinates because tracking target is not defined!")

    def scaleCoordsBuffer(self):
        self.points *= self.pix_size / self.map_buffer.pix_size
        self.geo_points *= self.pix_size / self.map_buffer.pix_size
        self.pix_size = self.map_buffer.pix_size

    def scaleCoordsModel(self):
        self.points *= self.pix_size / self.model.pix_size
        self.geo_points *= self.pix_size / self.model.pix_size
        self.pix_size = self.model.pix_size

    def getSSDistances(self):
        return np.linalg.norm(self.points_ss, axis=1)
    
    def checkISLimit(self, remove=False):
        """Checks if all points are within image shift limits."""

        for p, (point, dist) in enumerate(zip(self.points_stage, self.getSSDistances())):
            if dist > self.imaging_params.IS_limit:
                log(f"WARNING: Point {p + 1} {point} requires image shifts ({round(dist, 1)} microns) beyond the image shift limit ({self.imaging_params.IS_limit})!")
                if remove:
                    log(f"ERROR: Removing of points is not implemented yet!")
                    pass # TODO remove point from all lists
    
    def makeTrack(self, id, nav):
        super().makeTrack(id)
        log("ERROR: Making track is not fully implemented yet.")
        #TODO deal with nav changes

    def prepareForAcquisition(self, map_file, nav):
        """Prepares virtual maps and navigator for acquisition."""

        # Check if map_file contains area
        if len(map_file.stem.split("_A")) > 1:
            area_name = map_file.stem
            map_name = map_file.stem.split("_A")[0]
        else:
            map_name = area_name = map_file.stem

        # Get lamella montage nav_id for template
        template_id = nav.getIDfromNote(map_name + ".mrc")

        # Convert to buffer pix size
        self.scaleCoordsBuffer()

        for p, point in enumerate(self.points):
            # Make virtual map from montage in buffer
            virt_map_file = map_file.parent / (area_name + "_tgt_" + str(p + 1).zfill(3) + "_view.mrc")
            virt_map = np.flip(self.map_buffer.getCropImage(point, np.flip(self.imaging_params.cam_dims)), axis=0)       # crop image and flip y-axis
            log(f"DEBUG: Virtual map dimenstions: {virt_map.shape}")
            utils.writeMrc(virt_map_file, virt_map, self.map_buffer.pix_size)

            # Add map to navigator
            nav_id = nav.newMapFromImg(img_file=virt_map_file, template_id=template_id, coords=self.points_stage[p], label=str(p + 1).zfill(3), note=virt_map_file.name)

            # Tracking target nav adjustments
            if p == 0:
                nav.items[nav_id].changeAcquire(1)
                nav.items[nav_id].changeNote(area_name + "_tgts.txt")

        nav.newPointGroup(self.geo_points_stage, "geo", color_id=5, stage_z=nav.items[template_id].stage[2])    

    def getTargetsInfo(self, area_name):
        """Returns target area data in dicts."""

        targets = []
        for p, point in enumerate(self.points_stage):
            targets.append({
                "tsfile": area_name + "_ts_" + str(p + 1).zfill(3) + ".mrc", 
                "viewfile": area_name + "_tgt_" + str(p + 1).zfill(3) + "_view.mrc", 
                "SSX": self.points_ss[p][0], 
                "SSY": self.points_ss[p][1], 
                "stageX": point[0], 
                "stageY": point[1], 
                "SPACEscore": self.scores[p], 
                "skip": "False"})

        geo_points = []
        for g, geo_point in enumerate(self.geo_points_stage):
            geo_points.append({
                "SSX": self.geo_points_ss[g][0], 
                "SSY": self.geo_points_ss[g][1], 
                "stageX": geo_point[0], 
                "stageY": geo_point[1], 
            })

        return targets, geo_points, self.settings


#!/usr/bin/env python
# ===================================================================
# ScriptName:   SPACEtomo_functions
# Purpose:      Functions necessary to run SPACEtomo.
#               More information at http://github.com/eisfabian/PACEtomo
# Author:       Fabian Eisenstein
# Created:      2023/10/04
# Revision:     v1.1
# Last Change:  2024/04/10: check if tgts file already exists, added additional montage settings
#               2024/04/09: saved absolute path for virt maps in nav
#               2024/04/02: make inspected file when saving targets
#               2024/03/26: added check for GPU using torch
#               2024/03/25: fixes after Krios 3 test, added check for IM step already done (using WG_distance_threshold), fixed log management
#               2024/03/19: fixes after Krios 3 test
#               2024/03/08: outsourced target selection to updateQueue (also replaces queueSpaceRun)
#               2024/03/07: added import of more than one target area per lamella
#               2024/03/06: added check for stage is busy before WG montage, added check for no geo points before cleaning them up
#               2024/02/14: split IM routine to run lamella detection externally
#               2024/02/13: split collectWGMap from lamella detection
#               2023/12/12: fixed proximity removal, added version check
#               2023/12/11: added removal of lamellae due to close proximity
#               2023/11/28: fixes after Krios test
#               2023/11/15: added center padding for IM lamella detection and consider multiple hits, removed lamellae close to grid edge, removed geo points beyond IS limit, fixes and improvements to findOffset
#               2023/11/06: used IM lamella detection as confirmation
#               2023/11/04: fixes after Krios test
#               2023/10/31: fixes after Krios test
#               2023/10/27: added lamella option to select all classes, added ice check for tracking point, adjusted plotting option
#               2023/10/25: added nav file management
#               2023/10/23: added extra tracking
#               2023/10/18: added virtual map for targeting
#               2023/10/16: fixes after Krios test, removed polygon (listToSEMarray not found in external script)
#               2023/10/12: fixes after Krios test, added findOffset for more lamellae if lamella not found
#               2023/10/04: outsourcing of functions from main SPACEtomo script
# ===================================================================

import serialem as sem
import os
import glob
import copy
import json
from datetime import datetime
import numpy as np
from PIL import Image
Image.MAX_IMAGE_PIXELS = None                                   # removes limit for large images
from skimage import transform, exposure
import mrcfile
import SPACEtomo.modules.ext as space_ext
import SPACEtomo.modules.utils as utils
from SPACEtomo.modules.utils import log

versionSPACE = "1.2dev"

sem.SuppressReports()
sem.SetNewFileType(0)		                                    # set file type to mrc in case user changed default file type
CUR_DIR = sem.ReportDirectory()
SPACE_DIR = os.path.dirname(__file__)

# Check if torch and GPU are available
try:
    from torch.cuda import is_available
    if is_available():
        DEVICE = "cuda"
    else:
        DEVICE = "cpu"
except:
    DEVICE = "cpu"

##### WG functions #####

# Collect montage on new grid
""" MOVED TO modules.scope
def collectWGMap(grid_name, mic_params, montage_overlap):
    log("Collecting whole grid map...")

    # Check if WG map already exists
    note_id = sem.NavIndexWithNote(grid_name + ".mrc")
    if note_id > 0:
        map_id = note_id
        log("NOTE: Grid map already found in navigator. Skipping acquisiion.")
    else:
        sem.GoToImagingState(str(mic_params.WG_image_state))
        if sem.ReportLowDose()[0] > 0:
             log("WARNING: Whole grid montage image state is in Low Dose! This might result in unreasonably large grid maps.")
        sem.SetupFullMontage(montage_overlap, grid_name + ".mrc")

        # Make sure stage is ready (sometimes stage is busy error popped up when starting montage)
        while sem.ReportStageBusy():
            sem.Delay(1, "s")

        sem.M()
        map_id = sem.NewMap(0, grid_name + ".mrc")
        sem.CloseFile()

    sem.ChangeItemLabel(int(map_id), grid_name[:6])
    if len(grid_name) > 6:
         log("WARNING: Grid name is longer than 6 characters and will be truncated for labels in the Navigator.")
"""

# Save WG montage as rescaled image
""" MOVED TO modules.buf
def saveWGMap(grid_name, map_dir, mic_params, model):
    log("Loading whole grid map...")

    map_id = int(sem.NavIndexWithNote(grid_name + ".mrc"))
    sem.LoadOtherMap(map_id)

    # Load buffer
    log("Processing whole grid map...")
    buffer_orig, *_ = sem.ReportCurrentBuffer()
    img_prop = sem.ImageProperties(buffer_orig)
    mic_params.WG_pix_size = float(img_prop[4])

    # Rescale to model pixel size
    rescale_factor = model.pix_size / mic_params.WG_pix_size
    sem.ReduceImage(buffer_orig, rescale_factor)                # faster than tranform.rescale
    buffer = "A"
    montage = np.asarray(sem.bufferImage(buffer))

    # Save WG montage to file
    log("Saving whole grid map...")
    montage_img = Image.fromarray(montage.astype(np.uint8))
    montage_img.save(os.path.join(map_dir, grid_name + '_wg.png'))
    log(f"Saved at: {os.path.join(map_dir, grid_name + '_wg.png')}")
"""

# Load lamella detection results and draw nav points
""" SPLIT AND MOVED TO main script, buf.py, nav.py, mod_wg.py
def drawNavPoints(grid_name, map_dir, mic_params, model, exclude_classes=[], distance_threshold=5):
    # Read bboxes
    with open(os.path.join(map_dir, grid_name + "_boxes.json"), "r") as f:
        bboxes = json.load(f, object_hook=utils.revertArray)

    # Load map to draw points
    map_id = int(sem.NavIndexWithNote(grid_name + ".mrc"))
    sem.LoadOtherMap(map_id)
    buffer, *_ = sem.ReportCurrentBuffer()
    rescale_factor = model.pix_size / mic_params.WG_pix_size

    # Load all PL and FL nav points
    PL_nav_items = []
    FL_nav_items = []
    for b, bbox in enumerate(bboxes):
        # Check nav for preliminary lamella items
        nav_id = int(sem.NavIndexWithLabel("PL" + str(b + 1)))
        if nav_id > 0:
             item = sem.ReportOtherItem(nav_id)
             PL_nav_items.append({"id": nav_id, "coords": np.array([item[1], item[2]]), "bbox": bbox})

        # Check nav for final lamella items
        nav_id = int(sem.NavIndexWithLabel("FL" + str(b + 1)))
        if nav_id > 0:
             item = sem.ReportOtherItem(nav_id)
             FL_nav_items.append({"id": nav_id, "coords": np.array([item[1], item[2]]), "bbox": bbox})

    # If lamellae already detected return
    if len(PL_nav_items) > 0 or len(FL_nav_items) > 0:
        log("WARNING: Already found navigator items for detected lamellae. Skipping intermediate mag images.")

    # Generate nav item for each lamella
    nav_ids = []
    final_bboxes = []
    for b, bbox in enumerate(bboxes):
        #if bbox[4] == model.categories.index("broken") and not include_broken:
        if model.categories[bbox[4]] in exclude_classes:
            continue
        size = [bbox[2] - bbox[0], bbox[3] - bbox[1]]
        center = [bbox[0] + size[0] / 2, bbox[1] + size[1] / 2]

        nav_ids.append(int(sem.AddImagePosAsNavPoint(buffer, int(center[0] * rescale_factor), int(center[1] * rescale_factor), 0)))
        _, x_coord, y_coord, *_ = sem.ReportOtherItem(nav_ids[-1])

        # Check if coords are within grid limits (default is +/- 990, but can be changed in SerialEM property "StageLimits", this property can not be read by script)
        if abs(x_coord) > 990 or abs(y_coord) > 990:
            nav_ids.pop(-1)
            continue

        # Check if point already existed in final lamella items or preliminary lamella items and keep bbox (should not differ too much)
        if len(FL_nav_items) > 0:
             check_nav_items = FL_nav_items
        elif len(PL_nav_items) > 0:
             check_nav_items = PL_nav_items
        else:
             check_nav_items = []

        for item in check_nav_items:
            if np.linalg.norm(np.array([x_coord, y_coord]) - item["coords"]) < distance_threshold:
                sem.DeleteNavigatorItem(nav_ids[-1])
                nav_ids.pop(-1)
                nav_ids.append(item["id"])
                break          

        final_bboxes.append(bbox)
        sem.ChangeItemColor(nav_ids[-1], model.cat_nav_colors[int(bbox[4])])
        sem.ChangeItemNote(nav_ids[-1], model.categories[int(bbox[4])] + " (" + str(int(round(bbox[5] * 100, 0))) + "%)")
        sem.ChangeItemLabel(nav_ids[-1], "PL" + str(b + 1))

    return nav_ids, final_bboxes
"""

"""MOVED TO main script
def findOffset(grid_name, map_dir, lamella_ids, initial_bboxes, mic_params, model, keep_threshold=0.7, distance_threshold=5, external=False):
    # Check if findOffset was already run and skip
    IM_map_files = glob.glob(os.path.join(map_dir, grid_name + "_IM*_wg.png"))
    if len(IM_map_files) > 0:
         log("WARNING: Intermediate mag images were already recorded. Skipping...")
         return lamella_ids, initial_bboxes
    
    # Go to WG image state to get beam settings
    sem.GoToImagingState(str(mic_params.WG_image_state))
    sem.Delay(1, "s")       # Add delay for DM/GIF to catch up (DM crashes ocasionally on Krios 3)

    # Switch to IM mag depending on if mag index or mag value was given
    if mic_params.IM_mag_index < 50:
        sem.SetMagIndex(mic_params.IM_mag_index)
    else:
        sem.SetMag(mic_params.IM_mag_index)

    new_lamella_ids = []
    new_bboxes = []
    shifted = False
    for l, lamella_id in enumerate(lamella_ids):
        log(f"Moving to lamella {l + 1}...")
        sem.MoveToNavItem(lamella_id)
        sem.R()
        # Load buffer
        img_prop = sem.ImageProperties("A")
        mic_params.IM_pix_size = float(img_prop[4])

        # Rescale to model pixel size
        log("Rescaling map for lamella detection...")
        rescale_factor = model.pix_size / mic_params.IM_pix_size
        sem.ReduceImage("A", rescale_factor)                    # faster than tranform.rescale
        log("Rescaling intensity...")
        montage = exposure.rescale_intensity(np.asarray(sem.bufferImage("A")), out_range=(0, 255)).astype(np.uint8)

        # Save WG montage to file
        log("Saving rescaled map...")
        montage_img = Image.fromarray(montage)
        montage_img.save(os.path.join(map_dir, grid_name + "_IM" + str(l + 1) + "_wg.png"))   

        # Run findLamellae
        log("Finding lamella...")
        if not external:
            bboxes = model.findLamellae(map_dir, grid_name + "_IM" + str(l + 1), device=DEVICE)
        else:
            # Wait for boxes file to be written
            while not os.path.exists(os.path.join(map_dir, grid_name + "_IM" + str(l + 1) + "_boxes.json")):
                log("Waiting for external lamella detection...")
                sem.Delay(1, "s")
            # Read bboxes
            with open(os.path.join(map_dir, grid_name + "_IM" + str(l + 1) + "_boxes.json"), "r") as f:
                bboxes = json.load(f, object_hook=utils.revertArray)

        # Deal with more than one lamella per image
        bboxes = sorted(bboxes, key=lambda x: np.linalg.norm(np.array([(x[0] + x[2]) / 2,(x[1] + x[3]) / 2]) - np.array([model.sidelen // 2, model.sidelen // 2])))     # sort according to distance from center

        for bbox in bboxes:
            log(f"Lamella bounding box: {bbox}")
            log(f"Lamella was categorized as: {model.categories[int(bbox[4])]} ({round(bbox[5] * 100, 1)} %)")

            center = [(bbox[0] + bbox[2]) / 2,(bbox[1] + bbox[3]) / 2]

            # Add nav point
            new_lamella_ids.append(int(sem.AddImagePosAsNavPoint("B", int(center[0] * rescale_factor), int(center[1] * rescale_factor), 0)))

            new_bboxes.append(bbox)
                
            # Shift by marker if not done yet
            if not shifted:
                _, x_old, y_old, *_ = sem.ReportOtherItem(lamella_id)
                _, x_new, y_new, *_ = sem.ReportOtherItem(new_lamella_ids[-1])
                shift = np.array([x_new - x_old, y_new - y_old])
            
                log(f"Shifting items by {shift}")
                sem.ShiftItemsByMicrons(x_new - x_old, y_new - y_old)
                sem.DeleteNavigatorItem(new_lamella_ids[-1])
                new_lamella_ids[-1] = lamella_id

                shifted = True
            else:
                sem.ChangeItemColor(lamella_id, 5)
            sem.ChangeItemColor(new_lamella_ids[-1], model.cat_nav_colors[int(bbox[4])])
            sem.ChangeItemNote(new_lamella_ids[-1], model.categories[int(bbox[4])] + " (" + str(int(round(bbox[5] * 100, 0))) + "%)")
            sem.ChangeItemLabel(new_lamella_ids[-1], "FL" + str(l + 1))
                
        # Deal with situation where high confidence lamella was found in WG map but not in IM
        if len(bboxes) == 0:
            if initial_bboxes[l][5] > keep_threshold:
                new_lamella_ids.append(lamella_id)
                new_bboxes.append(initial_bboxes[l])
                log(f"No lamella detected in intermediate mag. Keeping original lamella due to high confidence ({initial_bboxes[l][5]})...")
            else:
                log("No lamella detected in intermediate mag. Removing lamella from list...")

    # Remove duplicates
    remove_ids = []
    lamella_coords = np.zeros((len(new_lamella_ids), 2))
    _, x, y, *_ = sem.ReportOtherItem(new_lamella_ids[0])
    lamella_coords[0, :] = x, y
    for lid1 in range(0, len(new_lamella_ids)):
        if lid1 in remove_ids: continue
        for lid2 in range(lid1 + 1, len(new_lamella_ids)):
            if lid2 in remove_ids: continue
            if np.array_equal(lamella_coords[lid2], np.zeros(2)):
                _, x, y, *_ = sem.ReportOtherItem(new_lamella_ids[lid2])
                lamella_coords[lid2, :] = x, y
            dist = np.linalg.norm(lamella_coords[lid2] - lamella_coords[lid1])
            if dist < distance_threshold:                       # if distance is smaller than 5 microns, remove lamella with smaller probability
                if new_bboxes[lid1][5] > new_bboxes[lid2][5]:
                    remove_ids.append(lid2)
                else:
                    remove_ids.append(lid1)
    if len(remove_ids) > 0:
        log(f"WARNING: Removed {len(remove_ids)} lamellae due to close proximity.")
        for rid in remove_ids:
            new_lamella_ids.pop(rid)
            new_bboxes.pop(rid)

    log(f"NOTE: {len(new_lamella_ids)} lamellae were confirmed in intermediate mag.")
    return new_lamella_ids, new_bboxes
"""

##### MM functions #####

# Collect new montage
""" MOVED TO scope.py
def collectMMMap(nav_id, map_name, mean_threshold, bbox, padding_factor, montage_overlap, mic_params, model):
    # Check if montage already exists
    note_id = int(sem.NavIndexWithNote(map_name + ".mrc"))
    if note_id > 0:
        map_id = note_id
        log("MM montage already available. Skipping acquisition...")
    else:
        # Determine montage size
        size = np.array([bbox[2] - bbox[0], bbox[3] - bbox[1]]) * padding_factor
        mont = np.round(size * model.pix_size / mic_params.view_pix_size / (mic_params.cam_dims - (min(mic_params.cam_dims) * montage_overlap))).astype(int) + np.ones(2, dtype=int)

        sem.MoveToNavItem(nav_id)

        # Check if targeting is totally off and allow for manual correction
        if mean_threshold > 0:
            sem.V()
            mean_counts = sem.ReportMeanCounts("A")
            if mean_counts < mean_threshold:
                user_input = sem.YesNoBox("\n".join(["DARK IMAGE", "", "The coordinates might be off target. Do you want to manually drag the image to a nearby lamella?"]))
                if user_input == 1:
                    sem.OKBox("\n".join(["Please center your target by dragging the image using the right mouse button! Press the <b> key when finished!", "Press <v> immediately following <b> if you want to take another View image."]))
                    log("NOTE: Please center your target by dragging the image using the right mouse button! Press the <b> key when finished!")
                    user_confirm = False
                    while not user_confirm:
                        while not sem.KeyBreak():
                            sem.Delay(0.1, "s")
                        user_confirm = True
                        for i in range(10):
                            if sem.KeyBreak("v"):
                                user_confirm = False
                                log("Take new view image!")
                                sem.V()
                                break
                            sem.Delay(0.1, "s")
                    sem.ResetImageShift()

        sem.Eucentricity(1)
        sem.UpdateItemZ(nav_id)
        
        log(f"Collecting medium mag montage at [{map_name.split("_")[-1]}] ({mont[0]}x{mont[1]})...")

        # Set montage parameters
        sem.ParamSetToUseForMontage(2)
        sem.SetMontPanelParams(1, 1, 1, 1)      # check all Montage control panel options
        sem.OpenNewMontage(mont[0], mont[1], os.path.join(CUR_DIR, map_name + ".mrc"))
        sem.SetMontageParams(1, int(montage_overlap * min(mic_params.cam_dims)), int(montage_overlap * min(mic_params.cam_dims)), int(mic_params.cam_dims[0]), int(mic_params.cam_dims[1]), 0, 1) # stage move, overlap X, overlap Y, piece size X, piece size Y, skip correlation, binning

        sem.M()
        map_id = int(sem.NewMap(0, map_name + ".mrc"))
        sem.ChangeItemLabel(map_id, "L" + map_name.split("_L")[-1])  
        sem.CloseFile()
    return map_id
"""

# Save montage as rescaled input image
""" MOVED TO buf.py
def saveMMMap(map_id, map_dir, map_name, mic_params, model, overwrite=False):
    if os.path.exists(os.path.join(map_dir, map_name + ".png")) and not overwrite:
        log("MM map already saved. Skipping saving!")
    else:
        sem.LoadOtherMap(map_id)
        buffer, *_ = sem.ReportCurrentBuffer()
        montage = np.asarray(sem.bufferImage(buffer))
        montage = transform.rescale(montage.astype(float) / 255, mic_params.view_pix_size / model.pix_size)
        montage = (montage * 255).astype(np.uint8)
        montage_img = Image.fromarray(montage)
        montage_img.save(os.path.join(map_dir, map_name + ".png"))
"""

# Save targets for PACEtomo   
""" MOVED TO main, buf.py, tgt.pt, nav.py
def saveAsTargets(buffer, map_dir, map_name, model, mic_params):
    # Check if tgts file for lamella already exists
    tgts_files = sorted(glob.glob(os.path.join(CUR_DIR, map_name + "*_tgts.txt")))
    if len(tgts_files) > 0:
         log(f"WARNING: Targets file for {map_name} already exists. Skipping target setup!")
         return

    # Load json data for all point files
    point_files = sorted(glob.glob(os.path.join(map_dir, map_name + "_points*.json")))
    target_areas = []
    for file in point_files:
        # Load json data
        with open(file, "r") as f:
            target_areas.append(json.load(f, object_hook=utils.revertArray))

    # Load map from buffer
    map_image = np.asarray(sem.bufferImage(buffer), dtype=float)

    # Loop over all found areas
    map_name_ori = map_name
    for ta, target_area in enumerate(target_areas):
        # Add area number to map_name if necessary
        if len(target_areas) > 1:
             map_name = map_name_ori + "_" + str(ta + 1)

        # Initiate lists
        targets = []
        nav_ids = []
        nav_virt_maps = []
        nav_virt_maps_stats = []
        geo_SS = []
        settings = {}
        if len(target_area["points"]) > 0:
            for i, point in enumerate(target_area["points"]):
                # Make virtual map
                virt_map_name = map_name + "_tgt_" + str(i + 1).zfill(3) + "_view.mrc"
                image_crop = np.flip(cropImage(map_image, (int(point[0] * model.pix_size / mic_params.view_pix_size), int(point[1] * model.pix_size / mic_params.view_pix_size)), mic_params.cam_dims[[1, 0]]), axis=0).astype(np.float32)
                nav_virt_maps_stats.append([np.min(image_crop), np.max(image_crop)])
                writeMrc(virt_map_name, image_crop, mic_params.view_pix_size)
                nav_virt_maps.append(virt_map_name)

                # Add nav item
                nav_ids.append(int(sem.AddImagePosAsNavPoint(buffer, int(point[1] * model.pix_size / mic_params.view_pix_size), int(point[0] * model.pix_size / mic_params.view_pix_size))))
                sem.ChangeItemNote(nav_ids[-1], virt_map_name)
                nav_info = sem.ReportOtherItem(nav_ids[-1])
                if i == 0:
                    stage0 = np.array([nav_info[1], nav_info[2]])
                    stage = stage0
                    SS = [0, 0]
                else:
                    stage = np.array([nav_info[1], nav_info[2]])
                    SS = mic_params.rec_s2ssMatrix @ (stage - stage0)

                # Check if target is within IS limits
                if SS[0] > mic_params.IS_limit or SS[1] > mic_params.IS_limit:
                    log(f"WARNING: Point {i + 1} requires image shifts ({round(SS[0], 1)}|{round(SS[1], 1)}) beyond the default image shift limit (15)!")

                # Add target
                targets.append({"tsfile": map_name + "_ts_" + str(i + 1).zfill(3) + ".mrc", "viewfile": map_name + "_tgt_" + str(i + 1).zfill(3) + "_view.mrc", "SSX": SS[0], "SSY": SS[1], "stageX": stage[0], "stageY": stage[1], "SPACEscore": target_area["scores"][i], "skip": "False"})

            # Get specimen coords of geo points
            for point in target_area["geo_points"]:
                # Add nav item
                geo_id = int(sem.AddImagePosAsNavPoint(buffer, int(point[1] * model.pix_size / mic_params.view_pix_size), int(point[0] * model.pix_size / mic_params.view_pix_size)))
                sem.ChangeItemColor(geo_id, 5)
                geo_item = sem.ReportOtherItem(geo_id)
                geo_coords = mic_params.rec_s2ssMatrix @ (np.array([geo_item[1], geo_item[2]]) - stage0)
                geo_SS.append({"SSX": geo_coords[0], "SSY": geo_coords[1]})

            # Set up nav for acquisition
            sem.ChangeItemNote(int(nav_ids[0]), map_name + "_tgts.txt")
            sem.SetItemAcquire(int(nav_ids[0]))

            # Change tgt points to virtual view maps
            changeNavPtsToMaps(nav_ids, nav_virt_maps, nav_virt_maps_stats)

            # Check if points file contains settings (use first occurence of settings, since they should be the same)
            if not settings and "settings" in target_area.keys():
                 settings = target_area["settings"]

        # Write PACE target file
        tgtsFilePath = os.path.join(CUR_DIR, map_name + "_tgts.txt")
        writeTargets(tgtsFilePath, targets, geo_SS, settings=settings)

        log(f"Target selection completed! {len(targets)} targets were selected.")

    # Make inspected file to prohibit user changing targets from GUI
    open(os.path.join(map_dir, map_name + "_inspected.txt"), "w").close()
"""

##### UTILITY FUNCTIONS #####

"""MOVED TO scope.py
class MicParams:
    def __init__(self, WG_image_state, IM_mag_index, MM_image_state, IS_limit):
        self.WG_image_state = WG_image_state
        self.IM_mag_index = IM_mag_index
        self.MM_image_state = MM_image_state

        self.IS_limit = IS_limit

        self.WG_pix_size = None
        self.IM_pix_size = None

    def getViewParams(self):
        cam_props = sem.CameraProperties()
        self.cam_dims = np.array([cam_props[0], cam_props[1]])
        self.view_pix_size = cam_props[4]   # [nm]

        self.view_c2ssMatrix = np.array(sem.CameraToSpecimenMatrix(0)).reshape((2, 2))
        self.view_ta_rotation = 90 - np.degrees(np.arctan(self.view_c2ssMatrix[0, 1] / self.view_c2ssMatrix[0, 0]))
        log(f"Tilt axis rotation: {self.view_ta_rotation}")
        self.view_rotM = np.array([[np.cos(np.radians(self.view_ta_rotation)), np.sin(np.radians(self.view_ta_rotation))], [-np.sin(np.radians(self.view_ta_rotation)), np.cos(np.radians(self.view_ta_rotation))]])

    def getRecParams(self):
        self.rec_s2ssMatrix = np.array(sem.StageToSpecimenMatrix(0)).reshape((2, 2))
        self.rec_beam_diameter = sem.ReportIlluminatedArea() * 100
        cam_props = sem.CameraProperties()
        self.rec_pix_size = cam_props[4]   # [nm]   

    def export(self, map_dir):
        with open(os.path.join(map_dir, "mic_params.json"), "w+") as f:
             json.dump(vars(self), f, indent=4, default=utils.convertArray)
        log(f"NOTE: Saved microscope parameters at {os.path.join(map_dir, 'mic_params.json')}")
"""

# Save settings for post action script
"""MOVED TO utils.py
def saveSettings(filename, vars, start="SEMflush", end="sem"):
    output = "# SPACEtomo settings from " + datetime.now().strftime("%d.%m.%Y %H:%M:%S") + "\n"
    save = False
    for var in vars:                                    # globals() is ordered by creation, start and end points might have to be adjusted if script changes
        if var == end:                                  # first var after settings vars
            break
        if save:
            if isinstance(vars[var], str):
                output += var + " = r'" + vars[var] + "'" + "\n"
            else:
                output += var + " = " + str(vars[var]) + "\n"
        if var == start:                                   # last var before settings vars
            save = True
    with open(filename + "_settings.txt", "w") as f:
        f.write(output)
"""
# Get grid name
""" MOVED TO scope.py
def getGridName(grid_slot, grid_default_name=""):
    # Check grid slot
    load_grid = False
    slot_status = sem.ReportSlotStatus(grid_slot)
    if grid_slot > 0 and slot_status[0] > 0:
        load_grid = True
        if len(slot_status) > 1 and isinstance(slot_status[-1], str):
            grid_name = slot_status[-1]
        else:
            grid_name = "G" + str(grid_slot).zfill(2)
    elif grid_slot == 0:
        if grid_default_name != "":
            grid_name = grid_default_name
        else:
            grid_name = "GX"
    else:
        sem.OKBox("Grid not found in slot. Please set grid_slot = 0 if it is already loaded!")
        sem.Exit()

    # Save old log before folder is changed (open empty log that will be discarded)
    sem.SaveLogOpenNew()

    return grid_name, load_grid
"""
# Load grid
""" MOVED TO scope.py
def loadGrid(grid_slot, grid_name, load_grid):
    # Open new navigator and log
    openNav(grid_name)
    sem.CloseLogOpenNew(1)      # Log was already saved in getGridName so force close here
    sem.SaveLog(0, grid_name)   # Set file name for new log

    log(f"SPACEtomo Version {versionSPACE}")
    sem.ProgramTimeStamps()

    if DEVICE == "cuda":
         log("NOTE: GPU will be used.")
    else:
         log("NOTE: No GPU found. CPU will be used.")

    # Initiate loading
    if load_grid:
        log(f"Loading grid [{grid_name}] from slot {grid_slot}...")
        sem.LoadCartridge(grid_slot)
"""
# Open new nav file if necessary
""" MOVED to nav.py
def openNav(grid_name):
    # Check if nav file is open
    nav_status = sem.ReportIfNavOpen()
    if nav_status < 2:
        if nav_status == 1:
            sem.SaveNavigator("temp.nav")
            sem.CloseNavigator()
            log("WARNING: Open navigator was saved as temp.nav and closed!")
        nav_file = os.path.join(CUR_DIR, grid_name + ".nav")
        sem.OpenNavigator(nav_file)
    else:
        # Check if current nav file is from loaded grid
        nav_file = sem.ReportNavFile()
        if os.path.splitext(os.path.basename(nav_file))[0] != grid_name:
            sem.SaveNavigator()
            nav_file = os.path.join(CUR_DIR, grid_name + ".nav")
            if os.path.exists(nav_file):
                sem.ReadNavFile(nav_file)
            else:
                sem.CloseNavigator()
                sem.OpenNavigator(nav_file)
"""
# Write PACEtomo target file
""" MOVED TO utils.py
def writeTargets(targetFile, targets, geoPoints=[], savedRun=False, resume={"sec": 0, "pos": 0}, settings={}):
    output = ""
    if settings != {}:
        for key, val in settings.items():
            if val != "":
                output += "_set " + key + " = " + str(val) + "\n"
        output += "\n"
    if resume["sec"] > 0 or resume["pos"] > 0:
        output += "_spos = " + str(resume["sec"]) + "," + str(resume["pos"]) + "\n" * 2
    for pos in range(len(targets)):
        output += "_tgt = " + str(pos + 1).zfill(3) + "\n"
        for key in targets[pos].keys():
            output += key + " = " + str(targets[pos][key]) + "\n"
        if savedRun:
            output += "_pbr" + "\n"
            for key in savedRun[pos][0].keys():
                output += key + " = " + str(savedRun[pos][0][key]) + "\n"
            output += "_nbr" + "\n"
            for key in savedRun[pos][1].keys():
                output += key + " = " + str(savedRun[pos][1][key]) + "\n"       
        output += "\n"
    for pos in range(len(geoPoints)):
        output += "_geo = " + str(pos + 1) + "\n"
        for key in geoPoints[pos].keys():
            output += key + " = " + str(geoPoints[pos][key]) + "\n"
        output += "\n"
    with open(targetFile, "w") as f:
        f.write(output)
"""

# Helper functions for virtual maps:

# Make sure nav contains template images
""" NOT NEEDED ANYMORE
def getTemplateID():
    view_id = int(sem.NavIndexWithNote("Template View"))
    if view_id == 0:
        sem.SetColumnOrGunValve(0)
        sem.AllowFileOverwrite(1)
        sem.OpenNewFile("template_view.mrc")
        sem.V()
        sem.S()
        view_id = int(sem.NewMap(0, "Template View"))
        sem.ChangeItemLabel(view_id, "TV")
        sem.CloseFile()
        sem.SetColumnOrGunValve(1)
        sem.AllowFileOverwrite(0)
    return view_id
"""
# Crop and pad the virtual map
""" MOVED TO buf.py
def cropImage(image, coords, fov):
	imageCrop = image[max(0, int(coords[0] - fov[0] / 2)):min(image.shape[0], int(coords[0] + fov[0] / 2)), max(0, int(coords[1] - fov[1] / 2)):min(image.shape[1], int(coords[1] + fov[1] / 2))]
	if not np.array_equal(imageCrop.shape, fov):
		mean = np.mean(imageCrop)
		if imageCrop.shape[0] < fov[0]:
			padding = np.full((int(fov[0] - imageCrop.shape[0]), imageCrop.shape[1]), mean)
			if int(coords[0] - fov[0] / 2) < 0:
				imageCrop = np.concatenate((padding, imageCrop), axis=0)
			if int(coords[0] + fov[0] / 2) > image.shape[0]:
				imageCrop = np.concatenate((imageCrop, padding), axis=0)
		if imageCrop.shape[1] < fov[1]:
			padding = np.full((imageCrop.shape[0], int(fov[1] - imageCrop.shape[1])), mean)
			if int(coords[1] - fov[1] / 2) < 0:
				imageCrop = np.concatenate((padding, imageCrop), axis=1)
			if int(coords[1] + fov[1] / 2) > image.shape[1]:
				imageCrop = np.concatenate((imageCrop, padding), axis=1)		
		log("WARNING: Target position is close to the edge of the map and was padded.")
	return imageCrop
"""
""" MOVED TO utils.py
def writeMrc(outfilename, image, pix_size):
	with mrcfile.new(os.path.join(CUR_DIR, outfilename), overwrite=True) as mrc:
		mrc.set_data(image)
		mrc.voxel_size = (pix_size * 10, pix_size * 10, pix_size * 10)
		mrc.update_header_from_data()
		mrc.update_header_stats()
"""
""" MOVED TO nav.py
def parseNav(navFile):
	with open(navFile) as f:
		navContent = f.readlines()
	header = []
	items = []
	newItem = {}
	index = 0

	for line in navContent:
		col = line.rstrip().split(" ")
		if col[0] == "": 
			if "Item" in newItem.keys():
				items.append(newItem)
				newItem = {}
				continue
			else:
				continue
		if line.startswith("[Item"):
			index += 1
			newItem = {"index": index, "Item": col[2].strip("]")}
		elif "Item" in newItem.keys():
			newItem[col[0]] = [val for val in col[2:]]
		else:
			header.append(line)
	if "Item" in newItem.keys():	#append last target
		items.append(newItem)
	return header, items

def writeNav(header, items, filename):
	text = ""
	for line in header:
		text += line
	text += os.linesep
	for item in items:
		text += "[Item = " + item["Item"] + "]" + os.linesep
		item.pop("Item")
		item.pop("index")
		for key, attr in item.items():
			text += key + " = "
			for val in attr:
				text += str(val) + " "
			text += os.linesep
		text += os.linesep
	with open(filename, "w", newline="") as f:
		f.write(text)
	return
"""
# Read nav file and change entries to map entries using the virtual maps
""" MOVED TO nav.py
def changeNavPtsToMaps(nav_ids, nav_maps, nav_virt_maps_stats):
    template_id = getTemplateID()   # make sure template exists before saving nav

    # Read nav file
    sem.SaveNavigator()
    nav_file = sem.ReportNavFile()
    nav_header, nav_items = parseNav(nav_file)

    # Make copy of view template entry
    template = copy.deepcopy(nav_items[template_id - 1])
    template.pop("RawStageXY")
    template.pop("SamePosId")

    # Determine image dimenstions for polygons
    ptsDX = np.array(np.array(template["PtsX"], dtype=float) - float(template["StageXYZ"][0]))
    ptsDY = np.array(np.array(template["PtsY"], dtype=float) - float(template["StageXYZ"][1]))    

    new_nav_ids = []
    for n, nav_id in enumerate(nav_ids):
        nav_file_id = nav_id - 1

        # Create new item with data from template and from nav point
        new_item = copy.deepcopy(template)
        new_item["Item"] = str(n + 1).zfill(3)
        new_item["StageXYZ"] = nav_items[nav_file_id]["StageXYZ"]
        new_item["Note"] = nav_items[nav_file_id]["Note"]
        if "Acquire" in nav_items[nav_file_id].keys(): 
            new_item["Acquire"] = nav_items[nav_file_id]["Acquire"]
        new_item["MapFile"] = [os.path.join(CUR_DIR, nav_maps[n])]
        new_item["MapMinMaxScale"] = nav_virt_maps_stats[n]    
        new_item["PtsX"] = ptsDX + float(nav_items[nav_file_id]["StageXYZ"][0])
        new_item["PtsY"] = ptsDY + float(nav_items[nav_file_id]["StageXYZ"][1])
    
        # Create new map ID
        unique_id = int(sem.GetUniqueNavID())
        while unique_id in new_nav_ids:
            unique_id = int(sem.GetUniqueNavID())
        new_nav_ids.append(unique_id)
        new_item["MapID"] = [unique_id]
        new_item["SamePosId"] = [unique_id]
    
        # Overwrite nav point with new nav item
        nav_items[nav_file_id] = copy.deepcopy(new_item)

    # Write nav file
    sem.SaveNavigator(nav_file)# + "~")						# backup unaltered nav file
    sem.CloseNavigator()
    writeNav(nav_header, nav_items, nav_file)
    sem.ReadNavFile(nav_file)
    
    log("Updated navigator file with target maps.")
"""


#!/usr/bin/env python
# ===================================================================
# ScriptName:   SPACEtomo_functions_ext
# Purpose:      Functions necessary to run SPACEtomo and that can be run externally without SerialEM.
#               More information at http://github.com/eisfabian/PACEtomo
# Author:       Fabian Eisenstein
# Created:      2023/10/04
# Revision:     v1.2dev
# Last Change:  2024/08/21: removed SPACE_DIR, fixed map staying in run list when error is detected
#               2024/07/31: allowed for empty seg to be 1x1 image
#               2024/07/25: added overlap when tiling in findLamellae function
#               2024/06/04: fixed findLamellae crash when img dims equal side lengths
#               2024/04/12: fixed weight_matrix dimension issue
#               2024/04/09: fixes after Rado Krios test, added PID output to SPACE_runs.json
#               2024/03/27: fixes after monitor test
#               2024/03/25: fixes after Krios 3 test, added check for truncated map in error file
#               2024/03/20: changed ultralytics import to be conditional to allow import of space_ext on K2 computer
#               2024/03/18: fixed extra tracking
#               2024/03/11: bug fixes and YOLO device control, added array conversion for json, got target selection working, adjusted plotTargets
#               2024/03/07: added TgtParams class to condense parameters for runTargetSelection function
#               2024/03/06: implemented clustering when targets spread over more area than image shift limits (still needs consideration in export)
#               2024/02/14: wrote general lamella detection function generally usable, fixed findGeoPoints crash when using lamella targeting
#               2024/02/13: split collectWGMap from lamella detection, split functions in to external functions file not requiring SerialEM
#               2023/12/12: fixed proximity removal, added version check
#               2023/12/11: added removal of lamellae due to close proximity
#               2023/11/28: fixes after Krios test
#               2023/11/15: added center padding for IM lamella detection and consider multiple hits, removed lamellae close to grid edge, removed geo points beyond IS limit, fixes and improvements to findOffset
#               2023/11/06: used IM lamella detection as confirmation
#               2023/11/04: fixes after Krios test
#               2023/10/31: fixes after Krios test
#               2023/10/27: added lamella option to select all classes, added ice check for tracking point, adjusted plotting option
#               2023/10/25: added nav file management
#               2023/10/23: added extra tracking
#               2023/10/18: added virtual map for targeting
#               2023/10/16: fixes after Krios test, removed polygon (listToSEMarray not found in external script)
#               2023/10/12: fixes after Krios test, added findOffset for more lamellae if lamella not found
#               2023/10/04: outsourcing of functions from main SPACEtomo script
# ===================================================================

import os
import sys
import glob
import json
import time
import subprocess
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, Ellipse
from PIL import Image
Image.MAX_IMAGE_PIXELS = None           # removes limit for large images
from scipy.ndimage import rotate
from scipy.spatial import distance_matrix
from scipy.optimize import minimize
from scipy.cluster.vq import kmeans2

import SPACEtomo.config as config
from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log


##### MM functions #####

class MMModel:
    def __init__(self, alt_model_dir=None):
        self.script = Path(__file__).parent.parent / config.MM_model_script

        # Allow for loading of specific model
        if alt_model_dir is None or not Path(alt_model_dir).exists():
            self.dir = Path(config.MM_model_folder)
        else:
            self.dir = Path(alt_model_dir)

        if not self.dir.exists() or not (self.dir / "dataset.json").exists():
            raise FileNotFoundError("Lamella segmentation model does not exist! Please import a model using the SPACEmodel command!")

        # Get all classes from model folder
        with open(self.dir / "dataset.json", "r") as f:
            dataset_json = json.load(f)
        self.categories = dataset_json["labels"]

        # Check json for pix size and fall back to config if not present
        if "pixel_size" in dataset_json.keys():
            self.pix_size = dataset_json["pixel_size"]
        else:
            self.pix_size = config.MM_model_pix_size

    def setDimensions(self, mic_params):
        self.beam_diameter = mic_params.rec_beam_diameter * 1000 / self.pix_size
        self.cam_dims = (mic_params.cam_dims[[1, 0]] * mic_params.rec_pix_size / self.pix_size).astype(int)    # record camera dimensions in pixels on view montage

    # Make empty segmentation to skip nnUnet and go to manual target selection
    @staticmethod
    def saveEmptySeg(map_dir, map_name):
        #map_image = Image.open(os.path.join(map_dir, map_name + ".png"))
        #seg = np.zeros(map_image.size, dtype=np.uint8)
        seg = np.zeros((1, 1), dtype=np.uint8)
        seg_image = Image.fromarray(seg)
        seg_image.save(os.path.join(map_dir, map_name + "_seg.png"))

# Class for lamella to enable target selection
class Lamella:
    #def __init__(self, map_name, map_dir, target_list, penalty_list, model, score_weights, score_weights_edge, grid_vecs, mic_params, max_tilt, plot=False):
    def __init__(self, map_name, map_dir, model, mic_params, tgt_params, plot=False, alt_seg_path=None):
        self.map_name = map_name
        self.map_dir = map_dir

        # Make subfolder for intermediary files
        self.out_dir = os.path.join(self.map_dir, self.map_name, "")
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        self.model = model
        self.mic_params = mic_params
        self.tgt_params = tgt_params

        self.plot = plot

        self.map = np.array(Image.open(os.path.join(map_dir, map_name + ".png")))

        # Load specific segmentation if specified or _seg suffix if not
        if alt_seg_path is not None and os.path.exists(alt_seg_path):
            self.segmentation = np.array(Image.open(alt_seg_path))
        else:
            self.segmentation = np.array(Image.open(os.path.join(map_dir, map_name + "_seg.png")))
        log("Generating target mask...")
        self.target_mask = self.makeMask(self.tgt_params.target_list)
        log("Generating background mask...")
        self.geo_mask = self.makeMask([name for name in self.model.categories if name not in self.tgt_params.target_list and name not in self.tgt_params.penalty_list])
        log("Generating penalty mask...")
        self.penalty_mask = self.makeMask(self.tgt_params.penalty_list)

        self.points = []
        self.point_scores = []
        self.clusters = []
        self.geo_points = []
        self.targets = []

    # Create mask from segmentation and selected classes
    def makeMask(self, class_names):
        mask = np.zeros(self.segmentation.shape)

        # Check if no class selected (sometimes happens for geo points when choosing whole lamella)
        if len(class_names) == 0:
            log("WARNING: Creating empty mask! This could cause an error later.")
            return mask

        for name in class_names:
            if name not in self.model.categories:
                log(f"ERROR: Unknown target class: {name}")
                log("Possible classes are: ")
                log(str(self.model.categories.keys()))
                break
            mask[self.segmentation == self.model.categories[name]] = 1

        # Save mask as binary image
        mask_img = Image.fromarray(mask, mode="1")
        mask_img.save(os.path.join(self.out_dir, self.map_name + "_mask_" + class_names[0] + ".png"))

        return mask

    # Define hexagonal grid of points depending on beam diameter
    def definePoints_grid(self, penalty_weight, threshold=0.0, alternative_mask=None):
        x1, x2 = self.tgt_params.vecs
        rec_dims = np.array(self.tgt_params.weight.shape)
        if alternative_mask is None:
            mask = self.target_mask
        else:
            mask = alternative_mask
        points = []
        point_scores = []
        max_rows = int(mask.shape[0] / abs(x1[0]) * 1.5)
        max_cols = int(mask.shape[1] / abs(x2[1]) * 1.5)
        for i in range(max_rows):
            for j in range(1, max_cols):
                point = i * x1 + j * x2 - max_rows // 2 * x1 - max_cols // 2 * x2 + np.array(mask.shape) // 2
            
                score = calcScore_point(point, mask, self.penalty_mask, penalty_weight, self.tgt_params.weight)
                if score > threshold:
                    points.append(point)
                    point_scores.append(score)
        return np.array(points), point_scores

    # Define sparse points depending on segmentation
    def definePoints_sparse(self, penalty_weight, threshold=0.0, alternative_mask=None):
        rec_dims = np.array(self.tgt_params.weight.shape)
        if alternative_mask is None:
            mask = self.target_mask
        else:
            mask = alternative_mask
        points = []
        point_scores = []
        i = rec_dims[0] / 2
        while i < mask.shape[0]:
            j = rec_dims[1] / 2
            while j < mask.shape[1]:
                point = np.array([i, j])
                score = calcScore_point(point, mask, self.penalty_mask, penalty_weight, self.tgt_params.weight)
                if score > threshold:
                    overlap = False
                    for pt in points:
                        dist = np.linalg.norm(pt - point)
                        if dist < self.model.beam_diameter:
                            overlap = True
                            break
                    if not overlap:
                        points.append(point)
                        point_scores.append(score)
                j += rec_dims[1]
            i += rec_dims[0]
        return np.array(points), point_scores
    
    # Add additional points surrounding existing points
    def addAdjacentPoints(self, penalty_weight, threshold=0.0):
        x1, x2 = self.tgt_params.vecs
        new_points = []
        new_point_scores = []
        for p, point in enumerate(self.points):
            new_points.append(point)
            new_point_scores.append(self.point_scores[p])
            for i in range(-1, 2):
                for j in range(-1, 2):
                    if i == j == 0:
                        continue
                    new_point = point + i * x1 + j * x2
                    for other_point in new_points:
                        dif_real = other_point - new_point
                        dif_squeezed = (self.mic_params.view_rotM.T @ dif_real) * np.array([1, np.cos(np.radians(self.tgt_params.max_tilt))])   # scale distance according to maximum tilt angle
                        if np.linalg.norm(dif_squeezed) < self.model.beam_diameter:
                            break  
                    score = calcScore_point(new_point, self.target_mask, self.penalty_mask, penalty_weight, self.tgt_params.weight)
                    if score > threshold:
                        new_points.append(new_point)
                        new_point_scores.append(score)
        log(f"Added {len(new_points) - len(self.points)} adjacent points!")
        log(f"{len(new_points)} points remaining.")
        self.points = np.array(new_points)
        self.point_scores = new_point_scores

    # Find clusters of points based on beam diameter
    def findClusters(self, points):
        if len(points) < 2:
            return [[0]]
        squeezed_points = np.array([(self.mic_params.view_rotM.T @ point) * np.array([1, np.cos(np.radians(self.tgt_params.max_tilt))]) for point in points])
        distM = distance_matrix(squeezed_points, squeezed_points)
        clusters = []
        for p1 in range(len(points)):
            clusters.append([p[0] for p in np.argwhere(distM[p1] <= 1.01 * self.model.beam_diameter)])
        
        clusters = sorted(clusters)
        
        for i in range(3):
            if len(clusters) > 1:
                final_clusters = [clusters[0]]
                for cluster in clusters[1:]:
                    cluster_set = set(cluster)
                    overlap = False
                    for fc in range(len(final_clusters)):
                        fcluster_set = set(final_clusters[fc])
                        if len(cluster_set & fcluster_set) != 0:
                            final_clusters[fc] = list(cluster_set | fcluster_set)
                            overlap = True
                            break
                    if not overlap:
                        final_clusters.append(cluster)  
                clusters = final_clusters
            
        log(f"Clusters found: {len(clusters)}")
        return clusters    

    # Clean points based on score threshold
    def cleanPoints_score(self, penalty_weight, threshold=0.0, clusters=None):
        cleaned_points = []
        cleaned_point_scores = []
        removed_point_ids = []
        for p, point in enumerate(self.points):
            score = calcScore_point(point, self.target_mask, self.penalty_mask, penalty_weight, self.tgt_params.weight)
            if score > threshold:
                cleaned_points.append(point)
                cleaned_point_scores.append(score)
            else:
                removed_point_ids.append(p)
        if clusters is not None:
            removed_point_ids = sorted(removed_point_ids, reverse=True)
            for p in removed_point_ids:
                cleaned_clusters = []
                for cluster in clusters:
                    cleaned_clusters.append([])
                    for i in cluster:
                        if i == p: continue
                        if i > p: cleaned_clusters[-1].append(i - 1)
                        else: cleaned_clusters[-1].append(i)
                clusters = [cluster for cluster in cleaned_clusters if cluster != []]
                
        log(f"Cleaned {len(removed_point_ids)} points by score threshold!")
        log(f"{len(cleaned_points)} points remaining.")
        self.points = np.array(cleaned_points)
        self.point_scores = cleaned_point_scores
        if clusters is not None:
            self.clusters = clusters

    # Clean points based on beam overlap (needs more flexible input and output to clean against other groups of points, e.g. geo points)
    def cleanPoints_dist(self, points, max_tilt=0, tolerance=0.1, point_scores=None, clusters=None):
        if point_scores is not None:    # sort points according to score to remove low score points first
            point_zip = sorted(zip(list(range(len(points))), points, point_scores), key=lambda x: x[2])
            points = [p for _, p, _ in point_zip]
            original_point_ids = [pid for pid, _, _ in point_zip]
        else:
            original_point_ids = list(range(len(points)))
        removed_point_ids = []
        moved_point_ids = []
        for p1 in range(len(points)):
            too_close = False
            for p2 in range(len(points)):
                if p1 == p2 or original_point_ids[p2] in removed_point_ids: continue
                dif_real = points[p2] - points[p1]
                dif_squeezed = (self.mic_params.view_rotM.T @ dif_real) * np.array([1, np.cos(np.radians(max_tilt))])   # scale distance according to maximum tilt angle
                if np.linalg.norm(dif_squeezed) < self.model.beam_diameter * (1 - 2 * tolerance * np.linalg.norm(dif_squeezed) / np.linalg.norm(dif_real)) or (np.linalg.norm(dif_squeezed) < 0.99 * self.model.beam_diameter and (p1 in moved_point_ids or p2 in moved_point_ids)):
                    too_close = True
                    break
                elif np.linalg.norm(dif_squeezed) < self.model.beam_diameter:
                    overlap = (self.model.beam_diameter - np.linalg.norm(dif_squeezed))
                    move = self.mic_params.view_rotM @ (overlap * dif_squeezed / np.linalg.norm(dif_squeezed) / np.array([1, np.cos(np.radians(max_tilt))]) / 2)
                    points[p1] = points[p1] - move
                    points[p2] = points[p2] + move
                    moved_point_ids.extend([p1, p2])
            if too_close:
                removed_point_ids.append(original_point_ids[p1])
        if point_scores is not None:
            point_zip = sorted(zip(original_point_ids, points), key=lambda x: x[0])
            points = [p for _, p in point_zip]
        cleaned_points = [point for p, point in enumerate(points) if p not in removed_point_ids]
        
        if clusters is not None:
            removed_point_ids = sorted(removed_point_ids, reverse=True)
            for p in removed_point_ids:
                cleaned_clusters = []
                for cluster in clusters:
                    cleaned_clusters.append([])
                    for i in cluster:
                        if i == p: continue
                        if i > p: cleaned_clusters[-1].append(i - 1)
                        else: cleaned_clusters[-1].append(i)
                clusters = [cluster for cluster in cleaned_clusters if cluster != []]
                
        log(f"Cleaned {len(removed_point_ids)} points by distance threshold!")
        log(f"{len(cleaned_points)} points remaining.")
        return np.array(cleaned_points), clusters
    
    # Remove single point from from beam overlap
    def distancePoint(self, point, points):
        moved = False
        for p in range(len(points)):
            dif_real = points[p] - point
            dif_squeezed = (self.mic_params.view_rotM.T @ dif_real) * np.array([1, np.cos(np.radians(self.tgt_params.max_tilt))])   # scale distance according to maximum tilt angle
            if np.linalg.norm(dif_squeezed) < self.model.beam_diameter:
                overlap = (self.model.beam_diameter - np.linalg.norm(dif_squeezed))
                move = self.mic_params.view_rotM @ (overlap * dif_squeezed / np.linalg.norm(dif_squeezed) / np.array([1, np.cos(np.radians(self.tgt_params.max_tilt))]))
                point = point - move         
                moved = True
        return point, moved
    
    # Create plot of targets on montage
    def plotTargets(self, offsets=[], clusters=[], tracking_id=None, overlay=None, save="temp.png"):
        rec_dims = np.array(self.tgt_params.weight.shape)
        if len(self.points) > 0:   
            if tracking_id is None:
                # Check if target list already exists and use first positions of each target area as tracking target
                if len(self.targets) > 0:
                    closest_middle = [points["points"][0] for points in self.targets]     # had to change all closes_middle to a list re accommodate more than one tracking target
                else:
                    middle_point = [np.max(self.points[:, 0]) / 2 + np.min(self.points[:, 0]) / 2, np.max(self.points[:, 1]) / 2 + np.min(self.points[:, 1]) / 2]
                    closest_middle = [self.points[np.argmin(np.linalg.norm(self.points - middle_point, axis=1))]]
            else:
                closest_middle = [self.points[tracking_id]]
        else:
            closest_middle = [np.array(self.map.shape) // 2]
        fig, ax = plt.subplots(figsize=(16, 12))
        
        if overlay is None:
            plt.imshow(self.map, cmap="gray")
        else:
            alpha = 1 / 4
            overlay_map = np.clip(np.dstack([self.map / 255 * (1 - alpha) + overlay * alpha, self.map / 255 * (1 - alpha), self.map / 255 * (1 - alpha)]), 0, 1)
            plt.imshow(overlay_map)
        plt.axline(closest_middle[0][[1, 0]], slope=-1 / np.tan(np.radians(self.mic_params.view_ta_rotation)), linestyle="--", color="#ffffff")
        
        if len(offsets) > 0 and len(clusters) > 1:
            for c, cluster in enumerate(clusters):
                data = np.array(self.points[cluster, :])
                plt.scatter(data[:, 1] + rec_dims[1] / 2, data[:, 0] + rec_dims[0] / 2, marker="*", s=100)
                offset_cluster_pixel = offsets[c] * max(rec_dims) * 10
                plt.arrow(self.points[cluster, :][0][1] - offset_cluster_pixel[1], self.points[cluster, :][0][0] - offset_cluster_pixel[0], offset_cluster_pixel[1], offset_cluster_pixel[0], length_includes_head=True, width=10, color="#ff7d00")
        
        for p, point in enumerate(self.points):
            ellipse = Ellipse((point[1], point[0]), self.model.beam_diameter / np.cos(np.radians(self.tgt_params.max_tilt)), self.model.beam_diameter, angle=self.mic_params.view_ta_rotation, fill=False, linewidth=1, color="#ffd700")
            ax.add_artist(ellipse)
            if np.any([np.array_equal(point, center) for center in closest_middle]):
                color = "#c92b27"
            else:
                color = "#578abf"
            rect = Rectangle([point[1] - rec_dims[1] / 2, point[0] - rec_dims[0] / 2], rec_dims[1], rec_dims[0], fill=False, linewidth=2, color=color)
            ax.add_artist(rect)
            plt.text(point[1] - 10, point[0] - 10, str(p + 1), color=color)

        if len(offsets) == 1:
            offset_pixel = offsets[0] * max(rec_dims) * 10
            plt.arrow(closest_middle[0][1] - offset_pixel[1], closest_middle[0][0] - offset_pixel[0], offset_pixel[1], offset_pixel[0], length_includes_head=True, width=10, color="#ff7d00")
        plt.axis("equal")
        plt.savefig(save, bbox_inches="tight", dpi=300)
        plt.clf()

    # Find targets based on mask
    def findPoints(self):
        # Precheck if mask is empty
        if not np.any(self.target_mask):
            return

        rec_dims = np.array(self.tgt_params.weight.shape)
        # Setup initial points
        if self.tgt_params.sparse:
            self.points, self.point_scores = self.definePoints_sparse(penalty_weight=0)
        else:
            self.points, self.point_scores = self.definePoints_grid(penalty_weight=0)

        log(f"Initial points found: {len(self.points)}")

        if self.plot: self.plotTargets(save=os.path.join(self.out_dir, self.map_name + "_points_initial.png"))

        if len(self.points) == 0:
            return False

        # Global translation optimization
        if not self.tgt_params.sparse:
            start_offset = np.zeros(2)
            offset = minimize(calcScore_cluster, start_offset, args=(self.points, self.target_mask, self.penalty_mask, 0.0, self.tgt_params.weight, self.tgt_params.edge_weights), method="nelder-mead", bounds=((-0.1, 0.1), (-0.1, 0.1)))
            offset_pixel = offset.x * max(rec_dims) * 10
            self.points += offset_pixel
            
            if self.plot: self.plotTargets(offsets=[offset.x], save=os.path.join(self.out_dir, self.map_name + "_points_globalSGD.png"))
            self.cleanPoints_score(penalty_weight=0, threshold=self.tgt_params.threshold)

        # Find clusters
        self.clusters = self.findClusters(self.points)

        # Iterative optimization
        if self.tgt_params.sparse: #len(self.score_weights_edge) > 0:
            adjacent_points = False
        else:
            adjacent_points = True
            
        for it in range(self.tgt_params.max_it):
            log("____________________")
            log(f"Iteration: {it + 1}")
            
            # Translation optimization for each cluster
            cluster_offsets = []
            total_offset = 0
            for cluster in self.clusters:
                start_offset = np.zeros(2)
                offset_cluster = minimize(calcScore_cluster, start_offset, args=(self.points[cluster, :], self.target_mask, self.penalty_mask, self.tgt_params.penalty, self.tgt_params.weight, self.tgt_params.edge_weights), method="nelder-mead", bounds=((-0.1, 0.1), (-0.1, 0.1)))
                offset_cluster_pixel = offset_cluster.x * max(rec_dims) * 10
                self.points[cluster, :] += offset_cluster_pixel
                cluster_offsets.append(offset_cluster.x)
                total_offset += np.linalg.norm(offset_cluster_pixel)

            if self.plot: self.plotTargets(offsets=cluster_offsets, clusters=self.clusters, save=os.path.join(self.out_dir, self.map_name + "_points_it" + str(it + 1).zfill(2) + ".png"))
            
            # Clean points
            prev_number = len(self.points)
            self.cleanPoints_score(self.tgt_params.penalty, self.tgt_params.threshold, self.clusters)
            self.points, self.clusters = self.cleanPoints_dist(self.points, self.tgt_params.max_tilt, tolerance=0.1, point_scores=self.point_scores, clusters=self.clusters)
            new_number = len(self.points)
            
            # Find new clusters
            if new_number != prev_number:
                if new_number == 0:
                    break
                self.clusters = self.findClusters(self.points)
            
            # Check for convergence
            log(f"Total offset: {total_offset}")
            if total_offset < 1 and new_number == prev_number:
                if not adjacent_points and it < self.tgt_params.max_it - 1:
                    log("Adding additional adjacent points...")
                    self.addAdjacentPoints(self.tgt_params.penalty)
                    self.clusters = self.findClusters(self.points)
                    adjacent_points = True
                if len(self.points) == prev_number:
                    log("##### Converged! #####\n")
                    break
        
        if len(self.points) > 0:
            # Calculate final SPACE scores
            for p, point in enumerate(self.points):
                self.point_scores[p] = calcScore_point(point, self.target_mask, self.penalty_mask, self.tgt_params.penalty, self.tgt_params.weight, self.tgt_params.edge_weights)

            # Check maximum distance
            coords_max = self.points.max(axis=0)
            coords_min = self.points.min(axis=0)
            max_delta = (coords_max - coords_min) * self.model.pix_size / 1000

            # Check distance to center point
            middle_point = [np.max(self.points[:, 0]) / 2 + np.min(self.points[:, 0]) / 2, np.max(self.points[:, 1]) / 2 + np.min(self.points[:, 1]) / 2]
            #print("middle", middle_point)
            closest_middle = self.points[np.argmin(np.linalg.norm(self.points - middle_point, axis=1))]
            #print("closest", closest_middle)
            furthest_middle = self.points[np.argmax(np.linalg.norm(self.points - closest_middle, axis=1))]
            #print("furthest", furthest_middle)
            middle_delta = np.abs(furthest_middle - closest_middle) * self.model.pix_size / 1000
            #print(middle_delta, max_delta)

            # If distance above threshold cluster targets in 2 clusters
            if max_delta[0] > 1.9 * self.mic_params.IS_limit or max_delta[1] > 1.9 * self.mic_params.IS_limit or middle_delta[0] > 0.95 * self.mic_params.IS_limit or middle_delta[1] > 0.95 * self.mic_params.IS_limit:
                log("Distance between points above threshold. Splitting targets in 2 acquisition areas...")
                centroids, label = kmeans2(self.points, 2, minit="++")
                counts = np.bincount(label)

                # Run up to 10 times in case one cluster remains empty
                for i in range(10):
                    if len(counts) > 1: break
                    centroids, label = kmeans2(self.points, 2, minit="++")
                    counts = np.bincount(label)    

                log(f"Split: {counts}")  
                self.clusters = [np.where(label == c)[0] for c in range(len(centroids))]

            else:
                self.clusters = [[i for i in range(len(self.points))]]

            log("Choosing tracking target...")

            # Consider tilted FOV to avoid ice coming in and compromising tracking
            expanded_dims = (self.mic_params.view_rotM @ ((self.mic_params.view_rotM.T @ np.array(self.tgt_params.weight.shape)) / np.array([1, np.cos(np.radians(self.tgt_params.max_tilt))]))).astype(int)

            self.targets = []       # list of array of targets for export as file
            # Choose tracking target for each cluster (in most cases it will be one)
            for cluster in self.clusters:
                points = self.points[cluster, :]

                # Find middle point, choose tracking target closest to it and move it to start of list of points
                middle_point = [np.max(points[:, 0]) / 2 + np.min(points[:, 0]) / 2, np.max(points[:, 1]) / 2 + np.min(points[:, 1]) / 2]

                if self.tgt_params.extra_track: # or len(self.points) == 1:
                    for it in range(3):
                        # Use middle point as initial tracking point and refine based on expanded dims and penalty mask only 
                        start_offset = (np.random.random(2) - 0.5) / 10   # add a random offset in case middle point coincides with only point
                        offset = minimize(calcScore_cluster, start_offset, args=([middle_point], np.zeros(self.target_mask.shape), self.penalty_mask, 1, np.ones(expanded_dims)), method="nelder-mead", bounds=((-0.1, 0.1), (-0.1, 0.1)))
                        offset_pixel = offset.x * max(rec_dims) * 10
                        middle_point += offset_pixel
                        # In case of overlap remove tracking point from other point
                        middle_point, moved = self.distancePoint(middle_point, points)
                        # No need for additional iterations if point was not moved after minimize
                        if not moved:
                            break
                    points = np.vstack([middle_point, points])

                    # Get the respective point scores
                    point_scores = np.array(self.point_scores)[cluster]
                    point_scores = np.insert(point_scores, 0, 0)
                    self.targets.append({"points": points, "scores": point_scores, "geo_points": []})
                    log("Successfully added additional tracking target.")
                else:                
                    closest_middle_ids = np.argsort(np.linalg.norm(points - middle_point, axis=1))
                    # Check 3 closest points to middle for ice
                    ice_scores = []
                    for min_id in range(0, min(3, len(closest_middle_ids))):
                        # Calculate score using expanded dims and only penalty mask
                        ice_scores.append(calcScore_point(points[closest_middle_ids[min_id]], np.zeros(self.target_mask.shape), self.penalty_mask, 1, np.ones(expanded_dims)))
                    log(f"Candidate ice scores: {ice_scores}")
                    # Use point with least ice for tracking
                    min_ice = np.argmax(ice_scores)

                    closest_middle = points[closest_middle_ids[min_ice]]
                    points = np.vstack([closest_middle, np.delete(points, closest_middle_ids[min_ice], 0)])

                    # Get the respective point scores
                    point_scores = np.array(self.point_scores)[cluster]
                    point_scores = np.concatenate([[point_scores[closest_middle_ids[min_ice]]], np.delete(point_scores, closest_middle_ids[min_ice], 0)])

                    self.targets.append({"points": points, "scores": point_scores, "geo_points": []})
                    log("Successfully chose tracking target.")


    # Find points to measure geometry
    def findGeoPoints(self):
        # Precheck is geo mask is empty
        if not np.any(self.geo_mask):
            return

        log("Finding geo points...")
        # Start from grid of points with high threshold
        geo_points, _ = self.definePoints_grid(penalty_weight=0.5, threshold=0.8, alternative_mask=self.geo_mask)

        if len(geo_points) > 0:
            # Remove points overlapping with target points at zero tilt
            all_points, _ = self.cleanPoints_dist(np.concatenate([geo_points, self.points]), max_tilt=0, tolerance=0.0)
            geo_points = all_points[:-len(self.points), :]

            for t, target_area in enumerate(self.targets):
                # Remove points beyond IS limits from tracking point
                area_geo_points = np.array([point for point in geo_points if np.linalg.norm(point - target_area["points"][0]) * self.model.pix_size < self.mic_params.IS_limit * 1000])
                if len(area_geo_points) < len(geo_points):
                    log(f"NOTE: Removed {len(geo_points) - len(area_geo_points)} geo points due to image shift limit.")

                # Only use 1 point per cluster if >=5 clusters
                if len(area_geo_points) > 5:
                    geo_clusters = self.findClusters(area_geo_points)
                    if len(geo_clusters) >= 5:
                        area_geo_points = area_geo_points[[cluster[0] for cluster in geo_clusters]]

                if len(area_geo_points) > 10:
                    # Choose 10 random geo points (TODO: update to use extreme points)
                    area_geo_points = np.random.default_rng().choice(area_geo_points, 10, replace=False)

                log(f"{len(area_geo_points)} geo points were selected.")

                if len(area_geo_points) < 3:
                    log("WARNING: Not enough geo points found to measure the sample geometry. The user defined pretilt and rotation values will be used.")
                #self.geo_points = geo_points
                self.targets[t]["geo_points"] = area_geo_points
        else:
            log("WARNING: Not enough geo points found to measure the sample geometry. The user defined pretilt and rotation values will be used.")

    # Export points
    def exportPoints(self):
        if len(self.targets) > 0:
            for t, target_area in enumerate(self.targets):
                #target_area["geo_points"] = self.geo_points
                #lamella = {"points": self.points, "point_scores": self.point_scores, "geo_points": self.geo_points}
                with open(os.path.join(self.map_dir, self.map_name + "_points" + str(t) + ".json"), "w+") as f:
                    json.dump(target_area, f, indent=4, default=utils.convertArray)
        else:
            # Write empty points file to ensure empty targets file is written and map is considered processed
            with open(os.path.join(self.map_dir, self.map_name + "_points.json"), "w+") as f:
                json.dump({"points": [], "scores": [], "geo_points": []}, f)            


# Function to instantiate lamella and run target selection from any script
def runTargetSelection(map_dir, map_name, tgt_params, mic_params, MM_model, save_plot=False, save_final_plot=False, alt_seg_path=None):
    map_name = os.path.basename(map_name)
    log(f"Setting up targets for {map_name}...")

    # Instantiate lamella and find points
    lamella = Lamella(map_name, map_dir, MM_model, mic_params, tgt_params, save_plot, alt_seg_path)
    lamella.findPoints()

    if len(lamella.points) == 0:
        log("WARNING: No targets found in automatic target selection!")
        # Write empty points file
        lamella.exportPoints()
        # Early return
        return
    
    log(f"Final targets: {len(lamella.points)}")
    if save_final_plot:
        log("Saving overview image...")
        lamella.plotTargets(overlay=lamella.target_mask, clusters=lamella.clusters, offsets=np.zeros([len(lamella.clusters), 2]), save=os.path.join(lamella.out_dir, map_name + "_" + tgt_params.target_list[0] +"_targets.png"))
        log(f"Saved at {os.path.join(lamella.out_dir, map_name + '_targets_' + tgt_params.target_list[0] + '.png')}")

    # Find geo points for sample geometry measurement
    lamella.findGeoPoints()

    # Export target points
    lamella.exportPoints()


##### UTILITY FUNCTIONS #####   

class MicParams_ext:
    def __init__(self, map_dir):
        with open(os.path.join(map_dir, "mic_params.json"), "r") as f:
            params = json.load(f, object_hook=utils.revertArray)
        for key, value in params.items():
            vars(self)[key] = value

    def export(self, map_dir):
        with open(os.path.join(map_dir, "mic_params.json"), "w+") as f:
             json.dump(vars(self), f, indent=4, default=utils.convertArray)
        log(f"NOTE: Saved microscope parameters at {os.path.join(map_dir, 'mic_params.json')}")

class TgtParams:
    def __init__(self, target_list=[], penalty_list=[], MM_model=None, sparse_targets=False, edge_targets=False, penalty_weight=0, score_threshold=0, max_tilt=0, mic_params=[], extra_tracking=False, max_iterations=10, file_dir=None):
        # Check if file can be imported
        if file_dir is None:
            self.target_list = target_list
            self.penalty_list = penalty_list

            # Replace keyword classes
            self.parseLists(MM_model)

            # Check if all lists classes are known to the model
            if not self.checkLists(MM_model):
                sys.exit()

            self.sparse = sparse_targets
            self.edge = edge_targets

            self.weight, self.edge_weights = makeScoreWeights(MM_model, self.edge)

            self.penalty = penalty_weight
            self.threshold = score_threshold

            self.max_tilt = max_tilt

            self.vecs = findGridVecs(MM_model, self.max_tilt, mic_params)

            self.extra_track = extra_tracking
            self.max_it = max_iterations
        else:
            self.loadFromFile(file_dir, MM_model)

    def parseLists(self, MM_model):
        # Replace lamella keyword
        if self.target_list[0] == "lamella":
            self.target_list = [name for name in MM_model.categories if name not in self.penalty_list]

        # Make sure penalty list includes black and white
        for guaranteed_penalty in ["black", "white", "dynabeads"]:
            if guaranteed_penalty in MM_model.categories and guaranteed_penalty not in self.penalty_list:
                log(f"WARNING: Added {guaranteed_penalty} class to avoid list!")
                self.penalty_list.append(guaranteed_penalty)

    def checkLists(self, MM_model):
        for cat in self.target_list + self.penalty_list:
            if cat not in MM_model.categories.keys() and cat not in ["lamella"]:
                log(f"ERROR: {cat} not found in model!")
                return False
        return True

    def loadFromFile(self, file_dir, MM_model):
        with open(os.path.join(file_dir, "tgt_params.json"), "r") as f:
            params = json.load(f, object_hook=utils.revertArray)
        for key, value in params.items():
            vars(self)[key] = value

        if self.weight is None:
            if MM_model is not None:
                self.weight, self.edge_weights = makeScoreWeights(MM_model, self.edge)
            else:
                log("ERROR: Need loaded MM model to regenerate weights!")
                return

    def export(self, map_dir):
        # Remove weights to make file smaller
        weight_matrix = self.weight
        weight_matrix_edge = self.edge_weights
        self.weight = None
        self.edge_weights = None

        # Write file
        with open(os.path.join(map_dir, "tgt_params.json"), "w+") as f:
             json.dump(vars(self), f, indent=4, default=utils.convertArray)

        # Restore weights
        self.weight = weight_matrix
        self.edge_weights = weight_matrix_edge

        log(f"NOTE: Saved target selection parameters at {os.path.join(map_dir, 'tgt_params.json')}")


# Find vectors for hexagonal grid
def findGridVecs(model, max_tilt, mic_params):
    rotM_pattern = np.array([[np.cos(np.radians(60)), -np.sin(np.radians(60))], [np.sin(np.radians(60)), np.cos(np.radians(60))]])
    x1 = np.array([model.beam_diameter, 0])
    x2 = rotM_pattern @ x1
    x2[1] /= np.cos(np.radians(max_tilt))

    x1 = mic_params.view_rotM @ x1
    x2 = mic_params.view_rotM @ x2
    if abs(mic_params.view_ta_rotation) > 45:
        xt = x1[:]
        x1 = x2[:]
        x2 = xt

    return np.vstack([x1, x2])

# Create score weights
def makeScoreWeights(model, edge=False, edge_angle_sampling=45):
    # Setup weight mask
    weight_mask = np.zeros(model.cam_dims)
    x, y = np.meshgrid(np.linspace(-model.cam_dims[1] // 2, model.cam_dims[1] // 2, model.cam_dims[1]), np.linspace(-model.cam_dims[0] // 2, model.cam_dims[0] // 2, model.cam_dims[0]))
    d = np.sqrt(x * x + y * y)
    sigma, mu = max(model.cam_dims) // 2, 0
    weight_mask = np.exp(-((d - mu) ** 2 / (2.0 * sigma ** 2)))
    weight_mask /= np.sum(weight_mask) / np.prod(model.cam_dims)

    # TODO add larger weight mask for penalty to account for ice chunks coming into field of view (needs max_tilt and boolean user setting?)
    # weight_mask_stretched = resize(weight_mask, [model.cam_dims[0] * np.cos(np.radians(max_tilt)), model.cam_dims[1]]) 

    # Setup edge masks
    edge_weight_masks = []
    if edge:
        x, y = np.meshgrid(np.linspace(-max(model.cam_dims), max(model.cam_dims), 2 * max(model.cam_dims)), np.linspace(-max(model.cam_dims), max(model.cam_dims), 2 * max(model.cam_dims)))
        d = np.sqrt(x * x + y * y)
        sigma, mu = max(model.cam_dims) // 2, 0
        square_weight = np.exp(-((d - mu) ** 2 / (2.0 * sigma ** 2)))
        square_weight /= np.sum(weight_mask) / np.prod(model.cam_dims) / 2
        
        edge_weight = np.array(square_weight)
        edge_weight[0:square_weight.shape[0] // 2, :] =  -square_weight[0:square_weight.shape[0] // 2, :]
        
        dif_shape = np.array(square_weight.shape) - np.array(weight_mask.shape)
        
        for i in range(0, 360, edge_angle_sampling):
            edge_weight_masks.append(rotate(edge_weight, angle=i, reshape=False)[dif_shape[0] // 2: dif_shape[0] // 2 + weight_mask.shape[0], dif_shape[1] // 2: dif_shape[1] // 2 + weight_mask.shape[1]])

    return weight_mask, edge_weight_masks

# Calculate score of a point depending on camera dimensions
def calcScore_point(point, mask, penalty_mask, penalty_weight, weight_mask, edge_weight_masks=[]):
    rec_dims = np.array(weight_mask.shape)
    if not rec_dims[0] / 2 <= point[0] < mask.shape[0] - rec_dims[0] / 2 or not rec_dims[1] / 2 <= point[1] < mask.shape[1] - rec_dims[1] / 2:
        return 0
    x_min = int(round(point[0] - rec_dims[0] / 2))
    y_min = int(round(point[1] - rec_dims[1] / 2))
    if len(edge_weight_masks) > 0:
        edge_scores = []
        for edge_mask in edge_weight_masks:
            score = np.sum(mask[x_min: x_min + edge_mask.shape[0], y_min: y_min + edge_mask.shape[1]] * edge_mask) / np.prod(rec_dims)
            score -= np.sum(penalty_mask[x_min: x_min + weight_mask.shape[0], y_min: y_min + weight_mask.shape[1]] * weight_mask) / np.prod(rec_dims) * penalty_weight
            edge_scores.append(score)
        score = np.max(edge_scores) 
    else:
        score = np.sum(mask[x_min: x_min + weight_mask.shape[0], y_min: y_min + weight_mask.shape[1]] * weight_mask) / np.prod(rec_dims)
        score -= np.sum(penalty_mask[x_min: x_min + weight_mask.shape[0], y_min: y_min + weight_mask.shape[1]] * weight_mask) / np.prod(rec_dims) * penalty_weight
    return score

# Calculate the total score of a cluster of points
def calcScore_cluster(offset, points, mask, penalty_mask, penalty_weight, weight_mask, edge_weight_masks=[]):
    rec_dims = np.array(weight_mask.shape)
    pixel_offset = offset * max(rec_dims) * 10
    total_score = 0
    for pt in points:
        point = pt + pixel_offset
        score = calcScore_point(point, mask, penalty_mask, penalty_weight, weight_mask, edge_weight_masks)
        total_score -= score
    return total_score


##### MONITOR FUNCTIONS #####

# Make list of map files in inventory
def monitorFiles(map_dir):
    file_list = sorted(glob.glob(os.path.join(map_dir, "*.png")))
    wg_list = []
    mm_list = []
    seg_list = []
    for file in file_list:
        if os.path.splitext(file)[0].split("_")[-1] == "wg":                # maps for lamella detection
            wg_list.append(os.path.splitext(os.path.basename(file))[0])
        elif os.path.splitext(file)[0].split("_")[-1] == "seg":             # segmentations
            seg_list.append(os.path.basename(file).rsplit("_seg", 1)[0])
        elif os.path.splitext(file)[0].split("_")[-1] == "segtemp":         # segmentation temp file
            continue
        else:
            mm_list.append(os.path.splitext(os.path.basename(file))[0])                       # all other maps
    # Check for unprocessed segmentation maps
    unprocessed_mm_list = []
    unprocessed_seg_list = []
    for map_name in mm_list:
        if map_name not in seg_list:
            unprocessed_mm_list.append(map_name)
        else:
            point_files = sorted(glob.glob(os.path.join(map_dir, map_name + "_points*.json")))
            if len(point_files) == 0:
                unprocessed_seg_list.append(map_name)

    # Check for unprocessed lamella detection maps
    unprocessed_wg_list = [wg_map for wg_map in wg_list if not os.path.exists(os.path.join(map_dir, wg_map + "_boxes.json"))]

    return unprocessed_mm_list, unprocessed_seg_list, unprocessed_wg_list 

def checkErr(map_dir, map_name):
    error_file = os.path.join(map_dir, map_name + "_SPACE.err")
    if os.path.exists(error_file):
        with open(error_file, "r", encoding="utf-8") as f:
            content = f.read()
            if "not enough memory" in content or "truncated" in content:
                log(f"ERROR: Out of memory or truncated segmentation error! Trying again...")
                return True
            elif "No such file or directory" in content:
                log(f"ERROR: File not found! Please ensure the config contains the proper segmentation script!")
                return True
            else:
                return False
    else:
        return False

# Control queue for external inference
def updateQueue(map_dir, WG_model=None, MM_model=None, mic_params=None, tgt_params=None, gpu_list=[0], save_plot=False):
    # Check run file
    if os.path.exists(os.path.join(map_dir, "SPACE_runs.json")):
        with open(os.path.join(map_dir, "SPACE_runs.json"), "r") as f:
            runs = json.load(f)
    else:
        runs = {"runs": [], "queue": []}

    # Get unprocessed maps
    mm_list, seg_list, wg_list  = monitorFiles(map_dir)

    # Remove finished runs
    for map_name in seg_list:
        run_list = [run["map"] for run in runs["runs"]]
        if map_name in run_list:
            runs["runs"].pop(run_list.index(map_name))

    # Determine free GPU
    free_gpus = []
    for gpu in gpu_list:
        if gpu not in [run["gpu"] for run in runs["runs"]]:
            free_gpus.append(gpu)

    # Run lamella detection
    if WG_model is not None:
        for wg_map in wg_list:
            # Check if GPU is available
            if len(free_gpus) > 0:
                device = "cuda:" + str(free_gpus[0])
            else:
                device = "cpu"
            # Check if file size is still changing
            file_size = os.path.getsize(os.path.join(map_dir, wg_map + ".png"))
            time.sleep(1)
            if file_size != os.path.getsize(os.path.join(map_dir, wg_map + ".png")):
                continue
            # Find lamellae
            log(f"Detecting lamellae on {os.path.basename(wg_map)}.png...")
            bboxes = WG_model.findLamellae(map_dir, os.path.basename(wg_map), suffix=".png", device=device)
            log(f"Detected {len(bboxes)} lamellae.\n")

    # Run target selection
    if tgt_params is not None:
        for seg_map in seg_list:
            runTargetSelection(map_dir, seg_map, tgt_params, mic_params, MM_model, save_plot)

    # Add unprocessed maps to queue if not already running
    add_to_queue = []   # add to queue after submitting new runs to give file time to finish copying
    for map_name in mm_list:
        if map_name in runs["queue"]:
            continue
        if map_name in [run["map"] for run in runs["runs"]]:
            if not checkErr(map_dir, map_name):
                continue
            else:
                # In case of detected error, remove map from run list and add to queue
                runs["runs"].pop(utils.findIndex(runs["runs"], "map", map_name))
                add_to_queue.append(map_name)
        else:
            add_to_queue.append(map_name)

    # Submit new run
    while len(free_gpus) > 0 and len(runs["queue"]) > 0:
        map_name = runs["queue"].pop(0)
        run_gpu = free_gpus.pop(0)

        out_file = open(os.path.join(map_dir, map_name + "_SPACE.err"), "w+")
        process = subprocess.Popen([sys.executable, Path(__file__).parent.parent / config.MM_model_script, os.path.join(map_dir, map_name + ".png")], env=dict(os.environ, CUDA_VISIBLE_DEVICES=str(run_gpu)), stdout=out_file, stderr=subprocess.STDOUT, text=True)
        log(f"\nStarting inference for {os.path.basename(map_name)}.png on GPU {run_gpu}...")
        runs["runs"].append({"map": map_name, "gpu": run_gpu, "PID": process.pid})

    # Add to queue after submitting new runs to give file time to finish copying
    for map_name in add_to_queue:
        runs["queue"].append(map_name)

    # Output
    if len(runs["runs"]) > 0:
        log("\n# Processing:")
        for run in runs["runs"]:
            log(f"{os.path.basename(run['map'])} (gpu: {run['gpu']})")

    if len(runs["queue"]) > 0:
        log("# Queue:")
        for map_name in runs["queue"]:
            log(os.path.basename(map_name))

    # Write run file
    with open(os.path.join(map_dir, "SPACE_runs.json"), "w+") as f:
        json.dump(runs, f, indent=4)


#!Python
# ===================================================================
#ScriptName SPACEtomo_postAction
# Purpose:      Post action script to setup targets from SPACEtomo runs.
#               More information at http://github.com/eisfabian/PACEtomo
# Author:       Fabian Eisenstein
# Created:      2023/07/13
# Revision:     v1.2
# Last Change:  2024/09/04: converted settings to configparser
#               2024/08/21: made callable
#               2024/08/20: adjusted to refactor
#               2024/04/10: fixed typo
#               2024/03/25: added start of acquisition if not already running
#               2024_03_18: added wait_for_inspection
#               2024/03/08: outsourced target selection to updateQueue, adjusted check for point files instead of segmentations
#               2024/03/07: read in mic_params file instead of instantiating it
#               2023/11/12: fixes after Krios test
#               2023/10/31: fixes after Krios test
#               2023/10/16: fixes after Krios test
#               2023/10/05: outsourced functions
#               2023/10/03: fixes after Krios test
#               2023/09/27: clean up
#               2023/09/26: added queue for SPACEruns
#               2023/09/22: switched to nnUnet
#               2023/07/13: copied and crudely adjusted target setup from SPACEtomo_WG
# ===================================================================

import sys
import time
from pathlib import Path
from SPACEtomo.modules.utils import log
import SPACEtomo.modules.ext as space_ext
import SPACEtomo.modules.utils as utils
from SPACEtomo.modules.scope import ImagingParams
from SPACEtomo.modules.nav import Navigator
from SPACEtomo.modules.mod_wg import WGModel
from SPACEtomo.modules.buf import Buffer
from SPACEtomo.modules.tgt import PACEArea

def main():
    CUR_DIR = utils.getCurDir()

    # Read settings written by main SPACE script
    settings = utils.loadSettings(CUR_DIR / "SPACEtomo_settings.ini")

    # Check if external processing directory is valid
    MM_external = False
    if settings["external_map_dir"] == "":
        MAP_DIR = CUR_DIR / "SPACE_maps"
    else:
        MAP_DIR = Path(settings["external_map_dir"])
        if MAP_DIR.exists():
            MM_external = True
        else:
            log(f"ERROR: External map directory [{MAP_DIR}] does not exist!")
            sys.exit()

    # Instantiate mic params from settings
    imaging_params = ImagingParams(None, None, None, CUR_DIR)

    # Instantiate nav
    nav = Navigator(is_open=True)

    # Load models
    WG_model = WGModel()
    MM_model = space_ext.MMModel()
    MM_model.setDimensions(imaging_params)

    # Instantiate tgt params from settings
    tgt_params = space_ext.TgtParams(file_dir=CUR_DIR, MM_model=MM_model)

    # Update SPACE runs and queue
    if not MM_external:
        space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params, tgt_params)

    # Check for unprocessed point files
    point_files = sorted(MAP_DIR.glob("*_points*.json"))
    tgt_files = sorted(CUR_DIR.glob("*tgts.txt"))

    unprocessed_point_files = []
    for point_file in point_files:
        map_name = point_file.name.split("_points")[0]
        if map_name not in [tgt_file.name.split("_tgts")[0] for tgt_file in tgt_files]:
            unprocessed_point_files.append(point_file)

    # Check for unprocessed maps
    mm_list, seg_list, wg_list  = space_ext.monitorFiles(MAP_DIR)

    # Check for areas still to be acquired
    num_acq = len(nav.searchByEntry("Acquire", ["1"]))

    # Check if collection would stop
    if num_acq == 0 and len(unprocessed_point_files) == 0:
        # Check if there will be further point files generated
        if len(mm_list) > 0 or len(seg_list) > 0:
            # Wait for another segmentation to be processed
            start_len_seg = len(seg_list)
            while len(seg_list) >= start_len_seg:
                log("Waiting for next prediction before setting up targets...")
                time.sleep(60)
                if not MM_external:
                    space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params, tgt_params)
                mm_list, seg_list, wg_list  = space_ext.monitorFiles(MAP_DIR)

            # Get point files that were not found previously
            point_files_new = sorted(MAP_DIR.glob("*_points*.json"))
            unprocessed_point_files = point_files_new - point_files

    # Set up targets for unprocessed point_files
    for point_file in unprocessed_point_files:
        map_name = point_file.name.split("_points")[0]

        # Check for map name areas to determine area_name
        map_areas = sorted(MAP_DIR.glob(map_name + "_points*.json"))
        if len(map_areas > 1):
            area_name = map_name + f"_A{map_areas.index(point_file) + 1}"
        else:
            area_name = map_name

        # Check if targets were inspected
        if settings["MM_wait_for_inspection"]:
            inspected = (MAP_DIR / (map_name + "_inspected.txt")).exists()
        else:
            inspected = True

        if not inspected:
            log("")
            log(f"WARNING: {map_name} targets are still waiting for inspection! Skipping...")
            continue

        log(f"\nSetting up targets for {map_name}...")

        # Load map from nav
        map_id = nav.getIDfromNote(map_name + ".mrc")
        map_img = Buffer(nav_id=map_id)

        # Instantiate target area
        target_area = PACEArea(point_file, MM_model, imaging_params, map_img)
        target_area.checkISLimit()

        # Set up navigator and create virtual maps
        target_area.prepareForAcquisition(CUR_DIR / (area_name + ".mrc"), nav)

        # Save target file
        targets, geo_points, tgt_settings = target_area.getTargetsInfo()
        utils.writeTargets(CUR_DIR / (area_name + "_tgts.txt"), targets, geo_points, settings=tgt_settings)

    num_acq = len(nav.searchByEntry("Acquire", ["1"]))
    log(f"Calling PACEtomo acquisition of {int(num_acq)} areas!")

    # Start acquisition if not already running
    #if not sem.ReportIfNavAcquiring()[0]:
    #    sem.StartNavAcquireAtEnd()


#!/usr/bin/env python
# ===================================================================
# ScriptName:   setup_model
# Purpose:      Download model and adjust config file.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/21
# Revision:     v1.2
# Last Change:  2024/09/24: added more model management (list, activate, remove, clear)
#               2024/09/04: fixed literal strings for model paths
# ===================================================================

import argparse
import shutil
from pathlib import Path
from urllib.request import urlretrieve
from urllib.parse import urlparse
from zipfile import ZipFile

from SPACEtomo.modules.utils import log, loadDatasetJson
from SPACEtomo import config

SPACE_DIR = Path(__file__).parent

ACCEPTED_EXT = [".pt", ".zip"]

def checkName(dir, name):
    """Checks if name already exists as model."""

    models = list(dir.glob(name + "*"))
    if models:
        log(f"ERROR: Model with name {models[0].stem} already exists! Please choose another name!")
        return False
    return True

def checkType(model_file):
    """Checks file extension to determine type of model."""

    if model_file.suffix == ".zip" or model_file.is_dir():
        model_type = "unet"
    elif model_file.suffix == ".pt":
        model_type = "yolo"
    else:
        log(f"ERROR: Currently models are only recognized as .pt or .zip files.")
        return False
    return model_type

def checkValidityUnet(model_dir):
    """Looks for essential files for model."""

    dataset_file = sorted(model_dir.glob("**/dataset.json"))
    folds = sorted(model_dir.glob("**/fold_*"))
    if not dataset_file or not folds:
        log(f"ERROR: {model_dir.name} is not a valid model and will be removed!")
        shutil.rmtree(model_dir)
        return False
    return dataset_file[0]

def cleanModelDir(model_dir):
    models = list(model_dir.glob("*"))
    for model in models:
        if model.name.endswith(".zip") or model.name == "unzip":
            model.unlink()
            log(f"WARNING: Removed failed model import [{model.name}]!")

def downloadModel(url, model_dir, model_name):
    """Downloads and saves model file."""

    url_path = Path(urlparse(url).path)

    # Check extension
    if url_path.suffix in ACCEPTED_EXT:

        # Check name
        if not model_name:
            model_name = url_path.stem
            if not checkName(model_dir, model_name):
                return False

        # Save model
        model_file = model_dir / (model_name + url_path.suffix)
        log(f"Downloading...")
        urlretrieve(url, model_file)
        return model_file
    else:
        log(f"ERROR: File type not known or not supported. Please provide the direct link to the model file {ACCEPTED_EXT}!")
        return False

def collectMetaDataYolo(file_name):
    """Collects meta data."""

    pix_size = None
    sidelen = None
    # Try to extract meta_data from filename
    cols = file_name.split("_")
    for col in cols:
        if "nm" in col:
            val = col.split("nm")[0]
            try:
                pix_size = float(val)
            except:
                pix_size = None
        elif "sl" in col:
            val = col.split("sl")[0]
            try:
                sidelen = float(val)
            except:
                sidelen = None

    meta_data = {}
    meta_data["WG_model_pix_size"] = pix_size
    meta_data["WG_model_sidelen"] = sidelen
    
    return meta_data

def collectMetaDataUnet(model_path):
    """Collects meta data."""

    meta_data = {}
    # Try to extract meta_data from dataset.json
    dataset_file = sorted(model_path.glob("**/dataset.json"))
    dataset_file = dataset_file[0]
    cats, pix_size, img_num = loadDatasetJson(dataset_file)
    folds = sorted(model_path.glob("**/fold_*"))
    meta_data["MM_model_folds"] = [fold.name.split("_")[-1] for fold in folds]
    meta_data["MM_model_pix_size"] = pix_size

    return meta_data

def listModels(model_dir):
    """Lists all models in models dir."""

    model_list = sorted(model_dir.glob("*"))
    log(f"\n############################################\n Registered SPACEtomo models: (* = active)")
    for model_file in model_list:
        if model_file.name.startswith("."):
            continue
        # Mark model as selected if config model path exists and is same as model_file
        if (not config.NO_WG_MODEL and Path(config.WG_model_file).samefile(model_file)) or (not config.NO_MM_MODEL and Path(config.MM_model_folder).samefile(model_file)):
            selected = "   *"
        else:
            selected = "    "
        log(f"{selected} {model_file.name}")
    if not model_list: 
        log(f"-")
    log("############################################\n")
    return

def activateModel(model_file, alternative_name=""):
    """Activates model by supplying path and meta_data to config."""
    
    # Check model type and validity
    model_type = checkType(model_file)
    if model_type == "unet" and not checkValidityUnet(model_file):
        return

    # Extract meta data
    if model_type == "yolo":
        name = alternative_name if alternative_name else model_file.name
        meta_data = collectMetaDataYolo(name)
    elif model_type == "unet":
        meta_data = collectMetaDataUnet(model_file)
    else:
        return

    # Adjust config
    with open(SPACE_DIR / "config.py", "r") as f:
        config_lines = f.readlines()

    new_config = []
    for line in config_lines:
        # YOLO
        if model_type == "yolo":
            # Model file name
            if line.strip().startswith("WG_model_file"):
                new_config.append(f"WG_model_file = \"{model_file.as_posix()}\"\n")
            # Pix size
            elif line.strip().startswith("WG_model_pix_size") and meta_data["WG_model_pix_size"]:
                new_config.append(f"WG_model_pix_size = {meta_data['WG_model_pix_size']} # nm/px\n")
            # Side length
            elif line.strip().startswith("WG_model_sidelen") and meta_data["WG_model_sidelen"]:
                new_config.append(f"WG_model_sidelen = '{meta_data['WG_model_sidelen']}'\n")
            else:
                new_config.append(line)

        # nnU-Net
        elif model_type == "unet":
            # Model folder
            if line.strip().startswith("MM_model_folder"):
                new_config.append(f"MM_model_folder = \"{model_file.as_posix()}\"\n")
            # Folds MM_model_folds = [0, 1, 2, 3, 4]
            elif line.strip().startswith("MM_model_folds") and meta_data["MM_model_folds"]:
                new_config.append(f"MM_model_folds = [{', '.join(meta_data['MM_model_folds'])}]\n")
            # Pix size
            elif line.strip().startswith("MM_model_pix_size") and meta_data["MM_model_pix_size"]:
                new_config.append(f"MM_model_pix_size = {meta_data['MM_model_pix_size']} # nm/px\n")
            else:
                new_config.append(line)

    with open(SPACE_DIR / "config.py", "w") as f:
        f.writelines(new_config)

    # Import new config values
    from SPACEtomo import config as config_new

    if model_type == "yolo":
        log(f"NOTE: YOLO model [{model_file.name}] was successfully activated.\nNOTE: Please double check the validity of the config values for:\nWG_model_pix_size = {config_new.WG_model_pix_size}   # nm/px\nWG_model_sidelen = {config_new.WG_model_sidelen}\nWG_model_categories = {config_new.WG_model_categories}")
    elif model_type == "unet":
        log(f"NOTE: Segmentation model [{model_file.name}] was successfully activated.\nNOTE: Please double check the validity of the config value for\nMM_model_pix_size = {config_new.MM_model_pix_size}   # nm/px")

def removeModel(model_file):
    """Deletes model and activates alternative if possible."""

    model_type = checkType(model_file)

    if model_file.is_dir():
        shutil.rmtree(model_file)
    else:
        model_file.unlink()
    log(f"NOTE: Model [{model_file.name}] was successfully removed!")

    model_list = sorted(model_file.parent.glob("*"))
    for model in model_list:
        if (model.is_dir() and model_type == "unet") or (model.suffix == ".pt" and model_type == "yolo"):
            activateModel(model)
            return

def addModel(model_dir, model_url, model_path, model_name=None):
    """Imports model file and unpacks it if necessary."""

    # Check if name exists
    if model_name:
        if not checkName(model_dir, model_name): 
            return

    # Check URL and download file
    if model_url:
        model_file = downloadModel(model_url, model_dir, model_name)
        if not model_file:
            return
        original_file_name = Path(urlparse(model_url).path).stem

        # Check name
        if not model_name:
            model_name = original_file_name

    # Check path and copy file
    elif model_path:
        file_path = Path(model_path)
        original_file_name = file_path.stem

        # Check extension
        if file_path.suffix in ACCEPTED_EXT:

            # Check name
            if not model_name:
                model_name = file_path.stem
                if not checkName(model_dir, model_name):
                    return

            # Copy model
            log(f"Importing...")
            model_file = model_dir / (model_name + file_path.suffix)
            shutil.copy(file_path, model_file)
        else:
            log(f"ERROR: File type not known or not supported. Please provide a path to the model file {ACCEPTED_EXT}!")
            return
    else:
        log("ERROR: Need either the URL of a model file to download or the path to a downloaded model file.")
        return

    if model_file.suffix == ".zip":
        # Create temp dir
        temp_dir = model_file.parent / "unzip"
        if temp_dir.exists():
            shutil.rmtree(temp_dir)
        temp_dir.mkdir()

        # Unzip model
        log(f"Extracting...")
        with ZipFile(model_file, "r") as zip_file:
            zip_file.extractall(temp_dir)
        
        # Delete zip file
        model_file.unlink()

        dataset_file = checkValidityUnet(temp_dir)
        if not dataset_file:
            return
        
        # Move to new model dir
        model_file = model_file.parent / model_name
        shutil.move(dataset_file.parent, model_file)
        temp_dir.rmdir()

    log(f"NOTE: New model {model_name} was imported!")

    activateModel(model_file, original_file_name)

def clearModels(model_dir):
    #TODO
    pass

def main():
    # Process arguments
    parser = argparse.ArgumentParser(description='Downloads or imports model and adjusts config file accordingly.')
    parser.add_argument('task', type=str, nargs="?", default='list', help='Task to be run: list (list models), activate (activate model), add (add a model), remove (remove a model by name) or clear (remove all models).')
    parser.add_argument('--url', dest='url', type=str, default=None, help='URL to trained YOLO model file [.pt] or zipped nnUnet model folder [.zip].')
    parser.add_argument('--path', dest='path', type=str, default=None, help='Path to downloaded trained YOLO model file [.pt] or zipped nnUnet model folder [.zip].')
    parser.add_argument('--name', dest='name', type=str, default=None, help='A unique name for the model.')
    args = parser.parse_args()

    # Make model folder
    model_dir = SPACE_DIR / "models"
    if not model_dir.exists():
        model_dir.mkdir()

    # Remove any non-model files from failed imports
    cleanModelDir(model_dir)

    # List models
    if args.task == "list":
        listModels(model_dir)
        return
    
    # Activate model
    if args.task == "activate":
        if args.name:
            if (model_dir / args.name).exists():
                activateModel(model_dir / args.name)
                return
            else:
                log(f"ERROR: Model with name [{args.name}] does not exist!")
                return
        else:
            log(f"ERROR: Please provide model name using --name!")
        
    # Add/import model and activate
    if args.task == "add":
        addModel(model_dir, args.url, args.path, args.name)
        return
    
    # Remove model
    if args.task == "remove":
        if args.name:
            if (model_dir / args.name).exists():
                removeModel(model_dir / args.name)
                return
            else:
                log(f"ERROR: Model with name [{args.name}] does not exist!")
        else:
            log(f"ERROR: Please provide model name using --name!")

    # Clear models
    if args.task == "clear":
        shutil.rmtree(model_dir)
        log(f"NOTE: All existing models were deleted. Please import new models to use SPACEtomo!")

if __name__ == "__main__":
    main()


#!/usr/bin/env python
# ===================================================================
# ScriptName:   SPACEtomo_nnUnet
# Purpose:      Runs a target selection deep learning model on medium mag montages of lamella and generates a segmentation that can be used for PACEtomo target selection.
#               More information at http://github.com/eisfabian/PACEtomo
# Author:       Fabian Eisenstein
# Created:      2023/05/19
# Revision:     v1.2
# Last Change:  2024/08/21: adjusted to refactor, made callable
#               2024/03/27: fixed device for lamella detection (again??)
#               2024/03/25: fixes after Krios 3 test
#               2024/03/11: Name fixes, added checkpoint file check
#               2024/02/15: added cpu inference if cuda not available (Patrick Cleeve)
#               2024/02/14: added import of SPACE ext functions
#               2023/11/06: changed to log output
#               2023/10/31: removed loading of settings and assume script is located inside SPACE_DIR
#               2023/10/18: fixed paths for external runs
#               2023/10/05: load config to get model data
#               2023/10/03: fixes after Krios test, padding to YOLO model side length	
#               2023/09/26: added YOLO bbox for lamella, read settings from settings file
#               2023/09/22: switched to nnUnet and simplified script
#               2023/07/10: more robust masking, binned heat maps, cleanup
#               2023/07/07: tested and fixes on Krios
#               2023/07/05: conversion to background process, implemented masked prediction speedup
#               2023/05/19: first test in SerialEM
# ===================================================================

import os
import sys
import time
import json
import logging
import numpy as np
from PIL import Image
Image.MAX_IMAGE_PIXELS = None
from pathlib import Path
import torch

import SPACEtomo.config as config
from SPACEtomo.modules.mod_wg import WGModel

def main():
    # Check if filename was given
    if len(sys.argv) == 2:
        montage_file = Path(sys.argv[1])
        if not montage_file.exists():
            print("ERROR: Map file does not exist!")
            sys.exit()
    else:
        print (f"Usage: python {sys.argv[0]} [input]")
        sys.exit("Missing arguments!")

    # Set directory to dir of given filename
    MAP_DIR = montage_file.parent

    # Set vars for nnUNet (not really used) before import to avoid errors
    os.environ["nnUNet_raw"] = str(MAP_DIR)
    os.environ["nnUNet_preprocessed"] = str(MAP_DIR)
    os.environ["nnUNet_results"] = str(MAP_DIR)

    from nnunetv2.inference import predict_from_raw_data as predict

    # Start log file
    logging.basicConfig(filename=MAP_DIR / (montage_file.stem + "_SPACE.log"), level=logging.INFO, format='')
    logging.info(f"Processing {montage_file}...")

    # Read segmentation classes
    with open(Path(config.MM_model_folder) / "dataset.json", "r") as f:
        dataset_json = json.load(f)
    classes = dataset_json["labels"]
    logging.info("Loaded class labels.")

    # Check device
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")   # dynamically set device
    is_cuda_device = True if device.type == "cuda" else False
    logging.info(f"Cuda available: {is_cuda_device}")

    # Load YOLO model
    WG_model = WGModel()

    # Load map
    start_time = time.time()
    montage_map = np.array(Image.open(montage_file))
    time_point1 = time.time()
    logging.info(f"Map was loaded in {int(time_point1 - start_time)} s.")

    # Rescale map for YOLO lamella bbox detection
    full_shape = np.array(montage_map.shape)
    division = np.round(full_shape * config.MM_model_pix_size / config.WG_model_pix_size).astype(int)
    pixel_shape = full_shape // division
    pixel_map = np.zeros(division)
    for i in range(division[0]):
        for j in range(division[1]):
            pixel = montage_map[i * pixel_shape[0]:(i + 1) * pixel_shape[0], j * pixel_shape[1]:(j + 1) * pixel_shape[1]]
            pixel_map[i, j] = np.mean(pixel)

    # Save rescaled map
    montage_img = Image.fromarray(pixel_map.astype(np.uint8))
    montage_img.save(MAP_DIR / (montage_file.stem + "_wg.png"))

    # Detect lamella
    bboxes = WG_model.findLamellae(MAP_DIR, montage_file.stem, save_boxes=True, device=device)

    # Check and upscale resulting box
    if len(bboxes) > 0:
        #bboxes = sorted(bboxes, key=lambda x: x[4], reverse=True)     # sort according to probability
        bboxes.sortBy("prob", reverse=True)                 # sort according to probability
        bbox = bboxes.boxes[0]

        # Scale coords up to full MMM size
        #bbox[0] = (bbox[0]) * pixel_shape[0]
        #bbox[2] = (bbox[2]) * pixel_shape[0]
        #bbox[1] = (bbox[1]) * pixel_shape[1]
        #bbox[3] = (bbox[3]) * pixel_shape[1]
        bbox = bbox * pixel_shape[0]        # pixel shape should always be square

        logging.info(f"Lamella bounding box: {bbox.xyxycc[:4]}")
        logging.info(f"Lamella was categorized as: {config.WG_model_categories[bbox.cat]} ({round(bbox.prob * 100, 1)} %)")

        bounds = np.round(bbox.xyxycc[:4]).astype(int)
        crop = montage_map[bounds[1]:bounds[3], bounds[0]:bounds[2]]

        time_point2 = time.time()
        logging.info(f"Bounding box was detected in {int(time_point2 - time_point1)} s.")
    else:
        bounds = None
        crop = montage_map

        time_point2 = time.time()
        logging.info(f"WARNING: No bounding box was detected in {int(time_point2 - time_point1)} s.")
        logging.info("Using whole montage map...")

    # Setup input and output for nnUNet
    dtype=np.float16 if is_cuda_device else np.float32                                      # half precision for cuda only
    input_img = np.array(crop, dtype=dtype)[np.newaxis, np.newaxis, :, :]

    # Check checkpoint files
    if (Path(config.MM_model_folder) / f"fold_{config.MM_model_folds[0]}" / "checkpoint_best.pth").exists():
        checkpoint_file = "checkpoint_best.pth"
    else:
        checkpoint_file = "checkpoint_final.pth"

    # Use temp name to pad later
    if bounds is None:
        out_name = MAP_DIR / (montage_file.stem + "_seg")
    else:
        out_name = MAP_DIR / (montage_file.stem + "_segtemp")

    # Do nnUNet inference
    predictor = predict.nnUNetPredictor(
        tile_step_size=0.5,
        perform_everything_on_device=is_cuda_device,
        device=device,
        allow_tqdm=False
    )

    predictor.initialize_from_trained_model_folder(
        config.MM_model_folder,
        config.MM_model_folds,
        checkpoint_name=checkpoint_file
    )

    logging.info("Model loaded.")
    logging.info("Starting prediction...")

    predictor.predict_single_npy_array(
        input_image=input_img, 
        image_properties={"spacing": [999, 1, 1]},
        output_file_truncated=str(out_name),
    )

    logging.info("Postprocessing...")

    # Pad segmentation to original map size
    if bounds is not None:
        segmentation = np.array(Image.open(str(out_name) + ".png"))
        padding = ((max(0, bounds[1]), max(0, full_shape[0] - bounds[3])), (max(0, bounds[0]), max(0, full_shape[1] - bounds[2])))
        segmentation = np.pad(segmentation, padding, constant_values=classes["black"])
        seg_out = Image.fromarray(np.uint8(segmentation))
        seg_out.save(MAP_DIR / (montage_file.stem + "_seg.png"))
        (out_name.parent / (out_name.stem + ".png")).unlink()

    logging.info(f"Prediction was completed in {round((time.time() - time_point2) / 60, 1)} min.")
    logging.info(f"Finished processing {montage_file.name}.")

if __name__ == "__main__":
    main()


#!/usr/bin/env python
# ===================================================================
# ScriptName:   CLI
# Purpose:      User interface for inspecting and labeling lamella grid atlases collected with SPACEtomo.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/04/26
# Revision:     v1.2
# Last Change:  2024/07/18: reworked GUI to class
# #             2024/06/07: added lamella table row highlighting, added save coords in YOLO format
#               2024/06/05: introduced Menu class, added shortcuts for most buttons
#               2024/06/04: added pix size check for export, added rescaling of mrc file
#               2024/06/03: added include empty tiles option, added save json option
#               2024/05/31: added status line, fixed tiles bounds, added next button, added hide/showElements
#               2024/05/30: added rescaling, saving as png, fixed export tiles
#               2024/05/29: outsourced plot clearing
#               2024/05/23: added lamella labels on plot, added padding when exporting edge tiles, fixed plotTiles
#               2024/05/10: added reading and writing of YOLO format labels, added saving of tiles with labels, added map class
#               2024/05/02: added lamella color from config, added reset zoom, delegated plot generation, call lamella detection
#               2024/04/30: added lamella focus buttons, outline scaling, list sorting
#               2024/04/26: Copy most of SPACEtomo_tgt
# ===================================================================

import sys
try:
    import dearpygui.dearpygui as dpg
except ModuleNotFoundError:
    print("WARNING: GUI library is not installed.")
import argparse
import shutil
from pathlib import Path

from SPACEtomo.modules.utils import log
from SPACEtomo.modules.gui.lam_sel import LamellaGUI
from SPACEtomo.modules.gui.tgt_sel import TargetGUI

# Attempt SerialEM import
import SPACEtomo.config as config
sys.path.insert(len(sys.path), config.SERIALEM_PYTHON_PATH)
try:
    import serialem as sem
    SERIALEM = True
except ModuleNotFoundError:
    SERIALEM = False

def show(gui, path=""):
    dpg.create_context()
    main = gui(path)
    main.show()
    dpg.destroy_context()
    
def run_process(process):
    log(f"Starting {process}...")
    process()
    
def main():
    # Process arguments
    parser = argparse.ArgumentParser(description="Command line interface for different SPACEtomo functions (e.g. GUIs).")
    parser.add_argument("task", type=str, default="targets", help="Name of the SPACEtomo task to run. Options: targets -> Open target selection GUI | lamella -> Open lamella selection GUI | run -> Start SPACEtomo run | scripts -> Get copy of SerialEM scripts")
    parser.add_argument("map_file", nargs="?", type=str, default="", help="Absolute path to a map file to load in GUI [.png].")
    args = parser.parse_args()

    # Get GUI
    gui = None
    process = None
    if args.task.lower() == "targets" or args.task.lower() == "target":
        gui = TargetGUI
    elif args.task.lower() == "lamella" or args.task.lower() == "lamellae":
        gui = LamellaGUI
    elif args.task.lower() == "run":
        if SERIALEM:
            from SPACEtomo import run
            process = run.main
        else:
            log(f"ERROR: SPACEtomo {args.task} requires SerialEM!")
    elif args.task.lower() == "scripts":
        serialem_script_dir = Path(__file__).parent / "SerialEM_scripts"
        target_path = Path.cwd() / "SerialEM_scripts"
        if not target_path.exists():
            shutil.copytree(serialem_script_dir, target_path)
            (target_path / "__init__.py").unlink()
            log(f"SerialEM scripts were successfully copied to: {target_path}")
        else:
            log(f"ERROR: A folder named SerialEM_scripts already exists!")
    else:
        raise ValueError(f"No task found with name {args.task}!")

    # Check map file
    if gui:
        map_file = ""
        if args.map_file != "": 
            map_file = Path(args.map_file)
            if not map_file.exists():
                map_file = ""
                log("WARNING: Map file does not exist! Attempting to start task without map...")            
    
        # Run GUI
        show(gui, map_file)
        
    if process:
        run_process(process)


if __name__ == "__main__":
    main()


#!Python
# ===================================================================
#ScriptName SPACEtomo_targetSelectionGUI
# Purpose:      Script to start SPACEtomo target selection GUI.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# ===================================================================

import serialem as sem
import sys
from SPACEtomo.modules.utils import log
try:
    import dearpygui.dearpygui as dpg
except:
    log("ERROR: DearPyGUI module not installed! If you cannot install it, please run the target selection GUI from an external machine.")
    sys.exit()
from pathlib import Path

from SPACEtomo.modules import utils
from SPACEtomo.modules.gui.tgt_sel import TargetGUI

def run(path=""):
    dpg.create_context()
    main = TargetGUI(path)
    main.show()
    dpg.destroy_context()

def main():
    # Look for SPACE_maps
    cur_dir = utils.getCurDir()
    space_maps = sorted(cur_dir.glob("**/SPACE_maps"))
    print(space_maps)
    if len(space_maps) > 0:
        run(space_maps[0])
    else:
        utils.log(f"ERROR: No SPACE_maps directory found within current directory [{cur_dir}]!")

if __name__ == "__main__":
    main()


#!/usr/bin/env python
# ===================================================================
# ScriptName:   buf
# Purpose:      Dummy interface for SerialEM buffer.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/23
# Revision:     v1.2
# Last Change:  2024/08/26: split from modules/buf.py
#               2024/08/23: added BufferDummy
# ===================================================================
 
import time
import json
import weakref
import numpy as np
from pathlib import Path
from PIL import Image
Image.MAX_IMAGE_PIXELS = None           # removes limit for large images
from skimage import transform, exposure
import mrcfile

from SPACEtomo import config
from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log

class BufferDummy:

    # Keep track of all instances to remove buffer letters when letters are overwritten
    instances = []

    # Get rolling buffers
    rolling_buffers = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K"]

    # Dummy info refs
    nav = None
    imaging_params = None

    # Keep file name of last image created by MicroscopeDummy
    last_dummy_image = None

    def __init__(self, buffer="A", nav_id=None, img=None) -> None:
        """Buffer can be initialized in three ways:
            1. Using a navigator ID of a map.
            2. Using a buffer containing an image.
            3. Using a buffer and an image to place into it."""

        # Initialize attributes
        self.img = None             # Image as numpy array
        self.img_mod = None         # Modified image as numpy array

        self.pix_size = None

        # Load nav map
        if nav_id is not None and self.nav:
            self.nav_id = nav_id
            self.buf, *_ = "A", None
            self.buf_mod = self.buf
            self.imgFromBuffer()

        # Use loaded buffer
        elif buffer and not img:
            self.nav_id = None
            self.buf = buffer
            self.buf_mod = self.buf

            if self.buf == "A" and self.last_dummy_image is not None:
                with mrcfile.open(self.last_dummy_image) as mrc:
                    self.img = np.flip(mrc.data, axis=0)
                    self.pix_size = mrc.voxel_size.x / 10

                log(f"#DUMMY: Loaded image from {self.last_dummy_image}")

            log(f"#DUMMY: Showing buffer {self.buf}!")
        
        # Use provided image
        elif buffer and img:
            self.nav_id = None
            self.buf = buffer
            self.buf_mod = self.buf
            self.img = img
            self.imgToBuffer()
            log(f"#DUMMY: Showing image in buffer {self.buf}!")

        else:
            raise ValueError("Need buffer letter (and optional image) or navigator item ID to initialize!")
        
        # Get image info
        if img:
            self.shape = np.array(img.shape)
        else:
            self.shape = self.imaging_params.cam_dims
        self.binning = 1
        if not self.pix_size:
            self.pix_size = 1

        # Get timestamp of image to make sure no other image was loaded in buffer when returning
        self.timestamp = time.time()

        # Add weak reference to instance list
        self.__class__.instances.append(weakref.ref(self))

    def show(self, reload=False):
        """Show buffer and assert that buffer is unchanged."""

        log(f"#DUMMY: Showing buffer {self.buf}!")

    def getMetadata(self, key):
        """Gets specific mdoc entry."""

        log(f"#DUMMY: Getting metadata entry from buffer for {key}. Returning 1.")

        return 1

    def imgFromBuffer(self, mod=False, reload=False):
        """Converts buffer to image as numpy array."""

        if not mod:
            if self.img is None or reload:
                if self.nav_id is not None:
                    map_file = self.nav.items[self.nav_id].entries["MapFile"][0]
                    with mrcfile.open(map_file) as mrc:
                        self.img = np.flip(mrc.data, axis=0)
                        self.pix_size = mrc.voxel_size.x / 10
                    log(f"#DUMMY: Loaded buffer image from {map_file}!")

                    log(f"DEBUG: Min val: {np.min(self.img)}, Max val: {np.max(self.img)}")
                else:
                    self.img = np.random.rand(*self.imaging_params.cam_dims)
                    log(f"#DUMMY: Generated random buffer image!")
        else:
            if self.img_mod is None or reload:
                self.img_mod = np.random.rand(*self.imaging_params.cam_dims)
                log(f"#DUMMY: Generated random modified buffer image!")

    def imgToBuffer(self, mod=False, buffer=None):
        """Shows img in buffer."""

        if not buffer:
            buffer = self.buf

        log(f"#DUMMY: Loaded image into buffer {buffer}!")

        # Roll buffers only if put in A (see script commands page for PutImageInBuffer)
        if buffer == "A":
            self.rollBuffers(skip=["A"])
            log(f"#DUMMY: Rolling buffers!")
        
    def saveImg(self, file_path, target_pix_size=None, save_meta_data=True, overwrite=False):
        """Saves buffer image as png image (optionally rescales it)."""

        # Check if file exists and handle overwriting
        if file_path.exists():
            if not overwrite:
                log(f"WARNING: {file_path.name} already exists! Skipping saving.")
                return
            else:
                log(f"WARNING: {file_path.name} already exists! It will be overwritten.")

        # Process image
        self.imgFromBuffer()

        if target_pix_size:
            # Rescale image
            rescale_factor = target_pix_size / self.pix_size
            self.rescaleImage(rescale_factor)
            img = self.img_mod  # Use modified image
        else:
            img = self.img  # Use unmodified image

        import concurrent.futures
        # Internal function that performs the actual save operation
        def save_operation(img):

            # Check intensity range and rescale if needed
            log(f"DEBUG: Image min/max: {np.min(img)}/{np.max(img)}")
            if np.min(img) < 0 or not 200 < np.max(img) < 255:
                log(f"DEBUG: Rescaling image intensity to out_range=(0, 255)")
                img = exposure.rescale_intensity(img, out_range=(0, 255)).astype(np.uint8)

            Image.fromarray(img).save(file_path)

            if save_meta_data:
                log(f"DEBUG: Saving meta data to {file_path.parent / (file_path.stem + '.json')}")
                self.saveMetaData(file_path=file_path.parent / (file_path.stem + ".json"), target_pix_size=target_pix_size)

        # Start the background thread to run `save_operation`
        executor = concurrent.futures.ThreadPoolExecutor()
        future = executor.submit(save_operation, img)


        """
        if file_path.exists():
            if not overwrite:
                log(f"WARNING: {file_path.name} already exists! Skipping saving.")
                return
            else:
                log(f"WARNING: {file_path.name} already exists! It will be overwritten.")

        self.imgFromBuffer()

        if target_pix_size:
            # Rescale image
            rescale_factor = target_pix_size / self.pix_size
            # Use rescale in case of upscaling
            self.rescaleImage(rescale_factor)
            # Use modified image
            img = self.img_mod
        else:
            # Use unmodified image
            img = self.img

        # Check intensity range and rescale if not between 0 and 255
        log(f"DEBUG: Image min/max: {np.min(img)}/{np.max(img)}")
        if np.min(img) < 0 or not 200 < np.max(img) < 255:
            log(f"DEBUG: Rescaling image intensity to out_range=(0, 255)") 
            img = exposure.rescale_intensity(img, out_range=(0, 255)).astype(np.uint8)

        Image.fromarray(img).save(file_path)

        """

        if save_meta_data:
            log(f"DEBUG: Saving meta data to {file_path.parent / (file_path.stem + '.json')}")
            self.saveMetaData(file_path=file_path.parent / (file_path.stem + ".json"), target_pix_size=target_pix_size)

        return future

    def saveMetaData(self, file_path, target_pix_size=None):
        """Saves meta data for buffer image as json file."""

        meta_data = {"datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()), "img_size": self.shape}

        if target_pix_size and target_pix_size != self.pix_size:
            meta_data["pix_size"] = target_pix_size
            meta_data["img_size"] = np.array(self.img_mod.shape)
            scaling_factor = target_pix_size / self.pix_size
        else:
            meta_data["pix_size"] = self.pix_size
            scaling_factor = 1

        # Get additional data from nav
        if self.nav and self.nav_id:
            meta_data["stage_coords"] = np.array(self.nav.items[self.nav_id].entries["StageXYZ"])
            if "MapScaleMat" in self.nav.items[self.nav_id].entries.keys():
                matrix = np.array([float(val) for val in self.nav.items[self.nav_id].entries["MapScaleMat"]]).reshape((2, 2))
                meta_data["s2img_matrix"] = matrix / scaling_factor
            if "MapFile" in self.nav.items[self.nav_id].entries.keys():
                meta_data["original_map"] = Path(" ".join(self.nav.items[self.nav_id].entries["MapFile"])).as_posix()
            if "MapTiltAngle" in self.nav.items[self.nav_id].entries.keys():
                meta_data["tilt_angle"] = float(self.nav.items[self.nav_id].entries["MapTiltAngle"][0])
            # Can be extended if necessary

        with open(file_path, "w") as f:
            json.dump(meta_data, f, indent=4, default=utils.convertArray)

    def rescaleImage(self, factor):
        """Rescales image using skimage (slower, but can also upscale)."""

        if self.img is not None:
            self.img_mod = (transform.rescale(self.img, 1 / factor) * 255).astype(np.uint8)
        else:
            raise ValueError("Image has not been converted from buffer yet!")

    def reduceBuffer(self, factor):
        """Downscales image using SerialEM's implementation of antialiased reduction with a Lanczos 2 filter."""

        self.rescaleImage(factor)
        self.buf_mod = "A"

        self.rollBuffers(skip=[self.buf_mod])

    def addNavPoint(self, point):
        """Adds buffer coords as nav point."""

        # Show buffer
        self.show()

        # Add point
        nav_id = self.nav.newPoint(str(len(self.nav)), np.random.rand(2) * 1800 - 900, "Dummy point")

        return nav_id

    def addNavBoxes(self, boxes, labels=[], padding_factor=1):
        """Adds polygons from boxes in image coords."""

        # Show buffer
        self.show()

        # Find scale factor
        if boxes.pix_size:
            factor = self.pix_size / boxes.pix_size
        else:
            factor = 1

        # Add boxes as polygon
        for b, box in enumerate(boxes.boxes):
            # Scale box coords
            box = box * factor

            # Check for stage limits (currently hardcoded as +/-990, TODO, once SerialEM property becomes readable by script command)
            stage_coords = self.px2stage(box.center) 
            if not (-990 < stage_coords[0] < 990 and -990 < stage_coords[1] < 990):
                continue

            # Pad box
            box.scale(padding_factor)

            # Create polygon points
            pts_x = [box.x_min, box.x_min, box.x_max, box.x_max, box.x_min]
            pts_y = [box.y_min, box.y_max, box.y_max, box.y_min, box.x_min]


            # Add polygon
            nav_id = self.nav.newPoint([*stage_coords, 0], box.label, "Dummy polygon")
            
            # Try to get stage rough stage coords
            if boxes.pix_size:
                factor = boxes.pix_size / 1000
            else:
                factor = 1

            self.nav.items[nav_id].entries["Type"] = ["1"]
            self.nav.items[nav_id].entries["PtsX"] = [str(x * factor) for x in pts_x]
            self.nav.items[nav_id].entries["PtsY"] = [str(y * factor) for y in pts_y]
            self.nav.items[nav_id].entries["NumPts"] = [str(len(pts_x))]

            # Adjust nav item
            self.nav.items[nav_id].entries["Color"] = [str(config.WG_model_nav_colors[box.cat])]

            # Choose label
            label = ""
            if labels and b < len(labels):
                label = labels[b]
            elif box.label:
                label = box.label
            
            if label:
                self.nav.items[nav_id].label = label
                self.nav.items[nav_id].note = f"{label}: {config.WG_model_categories[box.cat]} ({round(box.prob * 100)} %)"
            else:
                self.nav.items[nav_id].note = f"{config.WG_model_categories[box.cat]} ({round(box.prob * 100)} %)"

        self.nav.push()

    def px2stage(self, point):
        """Converts image coords to stage coords using buffer specific image and montage shifts."""

        if self.nav_id:
            stage = self.nav.items[self.nav_id].stage[:2]
            stage += point * self.pix_size / 1000
        else:
            stage = np.random.rand(2) * 1800 - 900

        return stage
       
    def stage2px(self, point):
        """Converts stage coords to image coords using buffer specific image and montage shifts."""

        return np.random.rand(2) * self.img.shape
    
    def rollBuffers(self, skip=[]):
        """Rolls buffers of all instances of Buffer."""

        # Check all instances
        for buf in self.instances:
            # Move buffer by 1 while in rolling buffers and remove if it moves beyond
            if buf.buf in self.rolling_buffers and buf.buf not in skip:
                buf.buf = chr(ord(buf.buf) + 1)
                if buf.buf not in self.rolling_buffers:
                    buf.buf = None

            # Do the same for the modified buffers
            if buf.buf_mod in self.rolling_buffers and buf.buf_mod not in skip:
                buf.buf_mod = chr(ord(buf.buf_mod) + 1)
                if buf.buf_mod not in self.rolling_buffers:
                    buf.buf_mod = None

    def getCropImage(self, coords, fov):
        """Crops out virtual map at given coordinates and field of view."""

        if self.img is None:
            self.imgFromBuffer()

        log(f"DEBUG:\nMap dimensions: {self.img.shape}\nCrop coords:\nx: {int(coords[0] - fov[0] / 2)}, {int(coords[0] + fov[0] / 2)}\ny: {int(coords[1] - fov[1] / 2)}, {int(coords[1] + fov[1] / 2)}")

        imageCrop = self.img[max(0, int(coords[0] - fov[0] / 2)):min(self.img.shape[0], int(coords[0] + fov[0] / 2)), max(0, int(coords[1] - fov[1] / 2)):min(self.img.shape[1], int(coords[1] + fov[1] / 2))]
        if not np.array_equal(imageCrop.shape, fov):
            mean = np.mean(imageCrop)
            if imageCrop.shape[0] < fov[0]:
                padding = np.full((int(fov[0] - imageCrop.shape[0]), imageCrop.shape[1]), mean)
                if int(coords[0] - fov[0] / 2) < 0:
                    imageCrop = np.concatenate((padding, imageCrop), axis=0)
                if int(coords[0] + fov[0] / 2) > self.img.shape[0]:
                    imageCrop = np.concatenate((imageCrop, padding), axis=0)
            if imageCrop.shape[1] < fov[1]:
                padding = np.full((imageCrop.shape[0], int(fov[1] - imageCrop.shape[1])), mean)
                if int(coords[1] - fov[1] / 2) < 0:
                    imageCrop = np.concatenate((padding, imageCrop), axis=1)
                if int(coords[1] + fov[1] / 2) > self.img.shape[1]:
                    imageCrop = np.concatenate((imageCrop, padding), axis=1)		
            log("WARNING: Target position is close to the edge of the map and was padded.")
        return imageCrop
    
    def __del__(self):
        """Cleans list of instances (but cannot delete current instance)."""

        remaining_refs = []
        for ref in self.__class__.instances:
            if ref():
                remaining_refs.append(ref)
        self.__class__.instances = remaining_refs